# .. Copyright (C) 2017 Bryan A. Jones.
#
#    This file is part of E-Book Binder.
#
#    E-Book Binder is free software: you can redistribute it and/or modify it
#    under the terms of the GNU General Public License as published by the Free
#    Software Foundation, either version 3 of the License, or (at your option)
#    any later version.
#
#    E-Book Binder is distributed in the hope that it will be useful, but WITHOUT
#    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#    details.
#
#    You should have received a copy of the GNU General Public License along
#    with E-Book Binder.  If not, see <http://www.gnu.org/licenses/>.
#
# .. highlight:: python
#
# ******************************************************************
# |docname| - Syntax highlighting for inline literal and code blocks
# ******************************************************************
# This is a heavily modified version of https://bitbucket.org/klorenz/sphinxcontrib-inlinesyntaxhighlight.
from docutils import nodes
import re
from sphinx.writers.html import HTMLTranslator

DIV_PRE_RE = re.compile(r"^<div[^>]*><pre>")
PRE_DIV_RE = re.compile(r"\s*</pre></div>\s*$")


def html_visit_literal(self, node):
    env = self.builder.env

    shall_highlight = False

    if (
        # This is a literal...
        (
            node["rawsource"].startswith("``")

            # ...that's not inside a ``:file:``
            and "role" not in node["attributes"]

            # ...and we should highlight literals, OR
            and env.config.inline_highlight_literals
        )

        # this is a code block, highlight.
        or ("code" in node["classes"])
    ):

        if env.config.inline_highlight_respect_highlight:
            lang = self.highlightlang
        else:
            lang = None

        highlight_args = node["get"]("highlight_args", {})

        if node["has_key"]("language"):
            # code-block directives
            lang = node["language"]
            highlight_args["force"] = True

        def warner(msg, **kwargs):
            self.builder.warn(self.builder.current_docname, msg, node["line"], **kwargs)

        highlighted = self.highlighter.highlight_block(
            node["astext"](), lang, warn=warner, **highlight_args
        )

        # highlighted comes as <div class="highlighted"><pre>...</pre></div>

        highlighted = DIV_PRE_RE.sub("", highlighted)
        highlighted = PRE_DIV_RE.sub("", highlighted)

        # import rpdb2 ; rpdb2.start_embedded_debugger('foo')

        starttag = self.starttag(
            node,
            "code",
            suffix="",
            CLASS="docutils literal highlight highlight-%s" % lang,
        )
        self.body.append(starttag + highlighted + "</code>")

    else:
        return old_html_visit_literal(self, node)

    raise nodes.SkipNode


old_html_visit_literal = HTMLTranslator.visit_literal
HTMLTranslator.visit_literal = html_visit_literal


def setup(app):
    app.add_config_value("inline_highlight_literals", True, "env")
    app.add_config_value("inline_highlight_respect_highlight", True, "env")



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__author__ = "isaiahmayerchak"

from docutils import nodes
from docutils.parsers.rst import directives
from runestone.server.componentdb import (
    addQuestionToDB,
    addHTMLToDB,
    maybeAddToAssignment,
)
from runestone.common.runestonedirective import RunestoneIdDirective, RunestoneIdNode
from runestone.common.xmlcommon import (
    write_substitute,
    substitute_visitor,
    substitute_departure,
)


def setup(app):
    app.add_directive("clickablearea", ClickableArea)

    app.add_node(
        ClickableAreaNode,
        html=(visit_ca_html, depart_ca_html),
        xml=(visit_ca_xml, depart_ca_xml),
    )

    app.add_config_value("clickable_div_class", "", "html")


TEMPLATE = """
<div class="runestone %(optclass)s">
<div data-component="clickablearea"  id="%(divid)s" data-question_label="%(question_label)s" %(optional)s %(table)s %(correct)s %(incorrect)s style="visibility: hidden;">
<span data-question>%(qnumber)s: %(question)s</span>%(feedback)s%(clickcode)s
"""
TEMPLATE_END = """
</div>
</div>
"""

XML_START = """
<exercise label="{divid}">
    <statement><p>{question}</p></statement>
"""


class ClickableAreaNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


# self for these functions is an instance of the writer class.  For example
# in html, self is sphinx.writers.html.SmartyPantsHTMLTranslator
# The node that is passed as a parameter is an instance of our node class.
def visit_ca_html(self, node):
    res = TEMPLATE

    node["delimiter"] = "_start__{}_".format(node["runestone_options"]["divid"])
    self.body.append(node["delimiter"])

    if "feedback" in node["runestone_options"]:
        node["runestone_options"]["feedback"] = (
            "<span data-feedback>" + node["runestone_options"]["feedback"] + "</span>"
        )
    else:
        node["runestone_options"]["feedback"] = ""

    if "iscode" not in node["runestone_options"]:
        node["runestone_options"]["correct"] = (
            "data-cc=" + '"' + node["runestone_options"]["correct"] + '"'
        )
        node["runestone_options"]["incorrect"] = (
            "data-ci=" + '"' + node["runestone_options"]["incorrect"] + '"'
        )
    else:
        node["runestone_options"]["correct"] = ""
        node["runestone_options"]["incorrect"] = ""

    res = res % node["runestone_options"]

    self.body.append(res)


def depart_ca_html(self, node):
    res = ""
    res = TEMPLATE_END % node["runestone_options"]
    self.body.append(res)

    addHTMLToDB(
        node["runestone_options"]["divid"],
        node["runestone_options"]["basecourse"],
        "".join(self.body[self.body.index(node["delimiter"]) + 1 :]),
    )

    self.body.remove(node["delimiter"])


def visit_ca_xml(self, node):
    res = XML_START.format(**node["runestone_options"])
    if node["runestone_options"].get("feedback", None):
        res += "<feedback><p>{feedback}</p></feedback>\n".format(
            **node["runestone_options"]
        )
    res += "<areas>\n"
    if "iscode" in node["runestone_options"]:
        # The case where iscode is not in options makes no sense and probably does not exist in
        # any runestone books
        for row in node["runestone_options"]["raw_source"]:
            row = row.replace("\n", "")
            row = row.replace(":click-correct:", "<area correct='yes'>")
            row = row.replace(
                ":click-incorrect:", "<area correct='no'>"
            )  # make explicit
            row = row.replace(":endclick:", "</area>")  # eliminate final colon as well
            row = "<cline>" + row + "</cline>\n"
            res += row
    self.output.append(res)


def depart_ca_xml(self, node):
    self.output.append("</areas></exercise>")


class ClickableArea(RunestoneIdDirective):
    """
    .. clickablearea:: identifier
        :question: Question text
        :feedback: Optional feedback for incorrect answer
        :iscode: Boolean that if present will put the content into a <pre>
        :table: Boolean that indicates that the content is a table.
        :correct: An array of the indices of the correct elements, separated by semicolons--if this is a table, each item in the array is a tuple
        with the first number being the row and the second number being the column--use a column number of 0 to make the whole row correct (ex: 1,2;3,0 makes the 2nd data cell in the first row correct as well as the entire 3rd row)
        :incorrect: An array of the indices of the incorrect elements--same format as the correct elements.

        --Content--


    config values (conf.py):

    - clickable_div_class - custom CSS class of the component's outermost div
    """

    required_arguments = 1
    optional_arguments = 0
    has_content = True
    final_argument_whitespace = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "question": directives.unchanged,
            "feedback": directives.unchanged,
            "iscode": directives.flag,
            "correct": directives.unchanged,
            "incorrect": directives.unchanged,
            "table": directives.flag,
        }
    )

    def run(self):
        """
        process the clickablearea directive and generate html for output.
        :param self:
        :return:
        .. clickablearea:: identifier
            :question: Question text
            :feedback: Optional feedback for incorrect answer
            :iscode: Boolean that if present will put the content into a <pre>
            :table: Boolean that indicates that the content is a table.
            :correct: An array of the indices of the correct elements, separated by semicolons--if this is a table, each item in the array is a tuple
            with the first number being the row and the second number being the column--use a column number of 0 to make the whole row correct (ex: 1,2;3,0 makes the 2nd data cell in the first row correct as well as the entire 3rd row)
            :incorrect: An array of the indices of the incorrect elements--same format as the correct elements.
            --Content--
        """
        super(ClickableArea, self).run()
        self.explain_text = [self.options["question"]]
        addQuestionToDB(self)

        self.assert_has_content()
        if "iscode" in self.options:
            source = "\n".join(self.content)
            source = source.replace(":click-correct:", "<span data-correct>")
            source = source.replace(":click-incorrect:", "<span data-incorrect>")
            source = source.replace(":endclick:", "</span>")
            source = "<pre>" + source + "</pre>"
            self.options["clickcode"] = source
            self.options["raw_source"] = self.content
        else:
            self.options["clickcode"] = ""
        clickNode = ClickableAreaNode()
        clickNode["runestone_options"] = self.options
        clickNode["source"], clickNode["line"] = self.state_machine.get_source_and_line(
            self.lineno
        )
        clickNode["template_start"] = TEMPLATE

        if "table" in self.options:
            self.options["table"] = "data-table"
        else:
            self.options["table"] = ""

        if "iscode" not in self.options:
            self.state.nested_parse(self.content, self.content_offset, clickNode)

        env = self.state.document.settings.env
        self.options["divclass"] = env.config.clickable_div_class

        maybeAddToAssignment(self)

        return [clickNode]



# *******************************
# |docname| - Add Questions to DB
# *******************************
#
#

# License
# -------
# Copyright (C) 2016-2020  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import print_function
import pdb

from datetime import datetime, timedelta

__author__ = "bmiller"

import os
from os import environ
import re
from sphinx.util import logging
from sqlalchemy import create_engine, Table, MetaData, select, and_, or_
from sqlalchemy.orm.session import sessionmaker

from runestone.common.runestonedirective import (
    RunestoneDirective,
    RunestoneIdNode,
)


logger = logging.getLogger(__name__)


def get_dburl(outer={}):
    """
    Return a nicely formatted database connection URL
    This function should not be used to configure a DAL db_uri for web2py that will already
    be configured in settings.

    :param outer:  pass locals from the calling environment
    :return:  string
    """
    # outer may contain the locals from the calling function
    # nonlocal env, settings # Python 3 only

    dburl = None
    if "SERVER_CONFIG" in environ:
        w2py_config = environ["SERVER_CONFIG"]
        if w2py_config == "development":
            dburl = environ["DEV_DBURL"]

        elif w2py_config == "production":
            dburl = environ["DBURL"]

        elif w2py_config == "test":
            dburl = environ["TEST_DBURL"]

    elif "options" in outer:
        dburl = outer["options"].build.template_args["dburl"]

    elif "env" in outer:
        dburl = outer["env"].config.html_context["dburl"]

    elif "env" in globals():
        dburl = globals()["env"].config.html_context["dburl"]

    elif "settings" in outer:
        dburl = outer["settings"].database_uri

    elif "settings" in globals():
        dburl = globals()["settings"].database_uri

    # We allow a dburl of "", hence the check against None.
    if dburl is not None:
        # DAL uses "postgres:", while SQLAlchemy (and the PostgreSQL spec) uses "postgresql:". Fix.
        remove_prefix = "postgres://"
        if dburl.startswith(remove_prefix):
            dburl = "postgresql://" + dburl[len(remove_prefix) :]
        return dburl

    raise RuntimeError("Cannot configure a Database URL!")


# Create a global DB query engine to share for the rest of the file
# We declare the variables here, but they don't get any values until setup is called.
# otherwise any import of runestone ends up trying to create database connections
# when it may not be necessary.
dburl = None
engine = None
meta = None
sess = None
questions = None
assignment_questions = None
courses = None
assignments = None
competency = None

# setup
# -----
#
# Here we connect two key functions to events in the build process.
# reset_questions and finalize updates
# this ensures that the database is in sync with the directives in the source
#

table_info = {}


def setup(app):
    global dburl, engine, meta, sess, questions, assignments, assignment_questions, courses, competency, chapters, sub_chapters, table_info

    app.connect("env-before-read-docs", reset_questions)
    app.connect("build-finished", finalize_updates)
    app.connect("env-before-read-docs", init_subs_file)
    app.connect("build-finished", finalize_subs_file)
    # the `qbank` option is for the QuestionBank
    # it allows us to populate the database from the question bank
    # but we don't care about populating chapter and subchapter tables and others
    # so we commit each question.
    app.add_config_value("qbank", False, "html")
    logger.info("Connecting to DB")
    try:
        dburl = get_dburl()
        # SQLite doesn't support ``client_encoding``, while PostgreSQL does.
        encoding = (
            dict(client_encoding="utf8") if dburl.startswith("postgresql") else {}
        )
        engine = create_engine(dburl, convert_unicode=True, **encoding)
        Session = sessionmaker()
        engine.connect()
        Session.configure(bind=engine)
        sess = Session()
    except Exception as e:  # psycopg2.OperationalError
        dburl = None
        engine = None
        sess = None
        print(e)
        print("Skipping all DB operations because environment variables not set up")
    else:
        # If no exceptions are raised, then set up the database.
        meta = MetaData()
        questions = Table("questions", meta, autoload=True, autoload_with=engine)
        assignments = Table("assignments", meta, autoload=True, autoload_with=engine)
        assignment_questions = Table(
            "assignment_questions", meta, autoload=True, autoload_with=engine
        )
        courses = Table("courses", meta, autoload=True, autoload_with=engine)
        competency = Table("competency", meta, autoload=True, autoload_with=engine)
        chapters = Table("chapters", meta, autoload=True, autoload_with=engine)
        sub_chapters = Table("sub_chapters", meta, autoload=True, autoload_with=engine)
        table_info = {
            "chapters": chapters,
            "sub_chapters": sub_chapters,
            "competency": competency,
            "courses": courses,
            "assignment_questions": assignment_questions,
            "assignments": assignments,
            "questions": questions,
        }


def get_engine_meta():
    return engine, meta, sess


def get_table_meta():
    return table_info


def init_subs_file(app, env, docnames):
    if app.builder.name == "xml":
        with open("rs-substitutes.xml", "w") as subfile:
            subfile.write("<substitutes>")


def finalize_subs_file(app, excpt):
    if app.builder.name == "xml":
        with open("rs-substitutes.xml", "a") as subfile:
            subfile.write("</substitutes>")


# Reset Questions
# ---------------
# We need to be able to tell the difference between questions that are
# part of the book and questions that are contributed by instructors
# so wheenver we build the book, we mark all questions for this base course
# with ``from_source = 'F'`` Then as the book is built and questions are updated
# those in the book get ``from_source = 'T'`` set.


def reset_questions(app, env, docnames):
    if sess:
        logger.info("Resetting questions")
        basecourse = env.config.html_context.get("basecourse")
        stmt = (
            questions.update()
            .where(
                and_(
                    questions.c.base_course == basecourse,
                    questions.c.question_type != "page",
                )
            )
            .values(from_source="F")
        )
        sess.execute(stmt)


# finalize updates
# ----------------
# If nothing has gone wrong during the build, then commit all the changes
# otherwise rollback to make sure we have everything consistent
def finalize_updates(app, excpt):

    if sess and excpt is None:
        if hasattr(app.env, "chap_titles"):
            try:
                update_chapter_subchapter(
                    app.env.chap_titles,
                    app.env.subchap_titles,
                    app.env.skips,
                    app.env.chap_numbers,
                    app.env.subchap_numbers,
                    app,
                )
                logger.info("Committing changes")
                sess.commit()
            except Exception as e:
                logger.error(f"Error while updating database -- details {e}")
                sess.rollback()
        else:
            logger.info("Skipping chapter DB update, no new information.")
    elif sess:
        logger.error("Rolling back database changes from build")
        sess.rollback()


def logSource(self):
    sourcelog = self.state.document.settings.env.config.html_context.get(
        "dsource", None
    )
    if sourcelog:
        with open(sourcelog, "a") as sl:
            sl.write("--------{}--------\n".format(self.arguments[0]))
            sl.write(".. {}:: {}\n".format(self.name, " ".join(self.arguments)))
            for k, v in self.options.items():
                sl.write("   :{}: {}\n".format(k, v))
            sl.write("   \n")
            sl.write("   " + "   \n".join(self.content))
            sl.write("\n")


def addQuestionToDB(self):
    # ``self`` must be a RunestoneDirective.
    assert isinstance(self, RunestoneDirective)
    question_id = False
    qbank = self.state.document.settings.env.config.qbank

    if dburl:
        if "basecourse" in self.options:
            basecourse = self.options["basecourse"]
        else:
            basecourse = self.state.document.settings.env.config.html_context.get(
                "basecourse", "unknown"
            )
        if basecourse == "unknown":
            raise self.severe("Cannot update database because basecourse is unknown")
            return

        last_changed = datetime.now()

        if "difficulty" in self.options:
            difficulty = self.options["difficulty"]
        else:
            difficulty = 3

        if "author" in self.options:
            author = self.options["author"]
        else:
            author = os.environ.get("USER", "Brad Miller")

        autograde = self.options.get("autograde", None)
        if "practice" in self.options:
            value = self.options["practice"].upper()
            assert (
                value == "T" or value == "F"
            ), f":practice: must be T or F not '{value}' for {self.options['divid']}"
            practice = value
        else:
            practice = None

        if ("topics" in self.options) and (self.options["topics"] != ""):
            topics = self.options["topics"]
        else:
            topics = "{}/{}".format(self.chapter, self.subchapter)
        #        topics = self.options.get('topics', "{}/{}".format(self.chapter, self.subchapter))
        if "data-optional" in self.options.get("optional", ""):
            optional = "T"
        else:
            optional = "F"

        from_source = self.options.get("from_source", "T")

        id_ = self.options["divid"]
        sel = select([questions]).where(
            and_(questions.c.name == id_, questions.c.base_course == basecourse)
        )
        res = sess.execute(sel).first()

        if not self.explain_text:
            self.explain_text = self.get_explain_text()
        et = " ".join(self.explain_text)[:80]

        # check for additional meta data to add
        meta_opts = {}
        if "pct_on_first" in self.options:
            meta_opts["pct_on_first"] = self.options["pct_on_first"]
        if "mean_clicks_to_correct" in self.options:
            meta_opts["mean_clicks_to_correct"] = self.options["mean_clicks_to_correct"]

        try:
            if res:
                stmt = (
                    questions.update()
                    .where(questions.c.id == res["id"])
                    .values(
                        question=self.block_text,
                        timestamp=last_changed,
                        is_private="F",
                        question_type=self.name,
                        subchapter=self.subchapter,
                        autograde=autograde,
                        author=author,
                        difficulty=round(float(difficulty)),
                        chapter=self.chapter,
                        practice=practice,
                        topic=topics,
                        from_source=from_source,
                        optional=optional,
                        description=et,
                        **meta_opts,
                    )
                )
                sess.execute(stmt)
                question_id = res["id"]
            else:
                ins = questions.insert().values(
                    base_course=basecourse,
                    name=id_,
                    question=self.block_text,
                    timestamp=last_changed,
                    is_private="F",
                    question_type=self.name,
                    subchapter=self.subchapter,
                    autograde=autograde,
                    author=author,
                    difficulty=round(float(difficulty)),
                    chapter=self.chapter,
                    practice=practice,
                    topic=topics,
                    from_source=from_source,
                    review_flag="F",
                    optional=optional,
                    description=et,
                    **meta_opts,
                )

                question_id = sess.execute(ins)
                question_id = question_id.inserted_primary_key[0]
            if qbank:
                sess.commit()
        except UnicodeEncodeError:
            raise self.severe(
                "Bad character in directive {} in {}/{}. This will not be saved to the DB".format(
                    id_, self.chapter, self.subchapter
                )
            )

        # check for prim_comp and supp_comp and add them to the competency table.
        prim_comps = []
        supp_comps = []
        if "prim_comp" in self.options:
            prim_comps = [
                x.strip() for x in self.options["prim_comp"].split(",") if x != ""
            ]
        if "supp_comp" in self.options:
            supp_comps = [
                x.strip() for x in self.options["supp_comp"].split(",") if x != ""
            ]

        if question_id:
            # Remove all competency info for this question,
            # then update it
            stmt2 = competency.delete().where(competency.c.question == question_id)
            sess.execute(stmt2)
            for comp in prim_comps:
                ins = competency.insert().values(
                    competency=comp,
                    question=question_id,
                    is_primary="T",
                    question_name=id_,
                )
                sess.execute(ins)
            for comp in supp_comps:
                ins = competency.insert().values(
                    competency=comp,
                    question=question_id,
                    is_primary="F",
                    question_name=id_,
                )
                sess.execute(ins)
            sess.commit()


def addQNumberToDB(app, node, qnumber):
    # ``self`` must be a RunestoneIdNode to contain a question number.
    assert isinstance(node, RunestoneIdNode)

    if not dburl:
        return

    basecourse = app.env.config.html_context.get("basecourse")
    if not basecourse:
        logger.error(
            "Cannot update database because basecourse is unknown.", location=node
        )
        return

    stmt = (
        questions.update()
        .where(
            and_(
                questions.c.name == node["runestone_options"]["divid"],
                questions.c.base_course == basecourse,
            )
        )
        .values(
            qnumber=qnumber,
        )
    )
    sess.execute(stmt)


def getQuestionID(base_course, name):

    sel = select([questions]).where(
        and_(questions.c.name == name, questions.c.base_course == base_course)
    )
    res = sess.execute(sel).first()
    if res:
        return res["id"]
    else:
        return None


def getAssignmentID(base_course, name):

    course = getCourseID(base_course)
    sel = select([assignments]).where(
        and_(assignments.c.name == name, assignments.c.course == course)
    )
    res = sess.execute(sel).first()
    if res:
        return res["id"]
    else:
        return None


def getOrInsertQuestionForPage(
    base_course=None,
    name=None,
    is_private="F",
    question_type="page",
    autograde="visited",
    author=None,
    difficulty=1,
    chapter=None,
):
    last_changed = datetime.now()

    sel = select([questions]).where(
        and_(questions.c.name == name, questions.c.base_course == base_course)
    )
    res = sess.execute(sel).first()

    if res:
        id = res["id"]
        stmt = (
            questions.update()
            .where(questions.c.id == id)
            .values(
                timestamp=last_changed,
                is_private=is_private,
                question_type=question_type,
                autograde=autograde,
                author=author,
                difficulty=difficulty,
                chapter=chapter,
            )
        )
        res = sess.execute(stmt)
        return id
    else:
        ins = questions.insert().values(
            base_course=base_course,
            name=name,
            timestamp=last_changed,
            is_private=is_private,
            question_type=question_type,
            autograde=autograde,
            author=author,
            difficulty=difficulty,
            chapter=chapter,
        )
        res = sess.execute(ins)
        return res.inserted_primary_key[0]


# Add a question to an Assignment
# -------------------------------
#
def addAssignmentQuestionToDB(
    question_name,
    assignment_name,
    points,
    basecourse,
    activities_required=0,
    autograde=None,
    which_to_grade=None,
    reading_assignment=None,
    sorting_priority=0,
):

    if not dburl:
        return

    question_id = getQuestionID(basecourse, question_name)
    assignment_id = getAssignmentID(basecourse, assignment_name)

    # now insert or update the assignment_questions row
    sel = select([assignment_questions]).where(
        and_(
            assignment_questions.c.assignment_id == assignment_id,
            assignment_questions.c.question_id == question_id,
        )
    )
    res = sess.execute(sel).first()
    vals = dict(
        assignment_id=assignment_id,
        question_id=question_id,
        activities_required=activities_required,
        points=points,
        autograde=autograde,
        which_to_grade=which_to_grade,
        reading_assignment=reading_assignment,
        sorting_priority=sorting_priority,
    )
    if res:
        # update
        stmt = (
            assignment_questions.update()
            .where(assignment_questions.c.id == res["id"])
            .values(**vals)
        )
        sess.execute(stmt)
    else:
        # insert
        ins = assignment_questions.insert().values(**vals)
        sess.execute(ins)


def maybeAddToAssignment(self):
    if self.in_exam:
        # this must come after this question has been added to the DB
        addAssignmentQuestionToDB(
            self.options["divid"],
            self.in_exam,  # assignment_name
            self.int_points,
            self.basecourse,
            autograde="pct_correct",
            which_to_grade="last_answer",
        )


def getCourseID(coursename):
    sel = select([courses]).where(courses.c.course_name == coursename)
    res = sess.execute(sel).first()
    return res["id"] if res else None


# Add an Assignment
# ------------------
# Historically we had two directives ``.. assignment::`` and ``.. usageassignment::``
# These allowed you to create assignments from the source for a book.
# However, these were deprecated in 2017. Version 3.0.4 is the most recent version
# prior to deprecation.  Recently (2020) work on the
# exam generator has renewed the need for similar functionality.
def addAssignmentToDB(
    name=None,
    course_name=None,
    deadline=None,
    points=0,
    is_timed="F",
    visible="F",
    time_limit=None,
):

    if not dburl:
        return

    course_id = getCourseID(course_name)
    sel = select([assignments]).where(
        and_(assignments.c.name == name, assignments.c.course == course_id)
    )
    res = sess.execute(sel).first()
    if deadline is None:
        deadline = datetime.now() + timedelta(weeks=1)
    if res:
        stmt = (
            assignments.update()
            .where(assignments.c.id == res["id"])
            .values(
                duedate=deadline,
                points=points,
                is_timed=is_timed,
                visible=visible,
                time_limit=time_limit,
                from_source="T",
            )
        )
        sess.execute(stmt)
        a_id = res["id"]
        # delete all existing AssignmentQuestions, so that you don't have any leftovers
        # this is safe because grades and comments are associated with div_ids and course_names, not assignment_questions rows.
        stmt2 = assignment_questions.delete().where(
            assignment_questions.c.assignment_id == a_id
        )
        sess.execute(stmt2)

    else:
        ins = assignments.insert().values(
            course=course_id,
            name=name,
            duedate=deadline,
            points=points,
            is_timed=is_timed,
            visible=visible,
            time_limit=time_limit,
            from_source="T",
            released="F",
        )
        res = sess.execute(ins)
        a_id = res.inserted_primary_key[0]

    return a_id


def addHTMLToDB(divid, basecourse, htmlsrc, feedback=None):
    if dburl:
        last_changed = datetime.now()
        sel = select([questions]).where(
            and_(questions.c.name == divid, questions.c.base_course == basecourse)
        )
        res = sess.execute(sel).first()
        try:
            if res:
                if res["htmlsrc"] != htmlsrc or res["feedback"] != feedback:
                    stmt = (
                        questions.update()
                        .where(questions.c.id == res["id"])
                        .values(
                            htmlsrc=htmlsrc, feedback=feedback, timestamp=last_changed
                        )
                    )
                    sess.execute(stmt)
        except UnicodeEncodeError:
            print("Bad character in directive {}".format(divid))
        except:
            print("Error while trying to add directive {} to the DB".format(divid))


def get_HTML_from_DB(divid, basecourse):
    sel = select([questions]).where(
        and_(questions.c.name == divid, questions.c.base_course == basecourse)
    )
    res = sess.execute(sel).first()
    if res:
        return res["htmlsrc"]
    else:
        return ""


def update_chapter_subchapter(
    chaptitles, subtitles, skips, chap_numbers, subchap_numbers, app
):
    """
    When the build is completely finished output the information gathered about
    chapters and subchapters into the database.
    """

    engine, meta, sess = get_engine_meta()
    table_info = get_table_meta()

    if not sess:
        logger.info("You need to install a DBAPI module - psycopg2 for Postgres")
        logger.info("Or perhaps you have not set your DBURL environment variable")
        return

    chapters = table_info["chapters"]
    sub_chapters = table_info["sub_chapters"]
    questions = table_info["questions"]

    basecourse = app.config.html_context.get("basecourse", "unknown")

    # Remove the chapter / subchapter info
    logger.info("Deleteing Chapter / Subchapter info for {}".format(basecourse))
    logger.info(
        "Warning - always use build --all to ensure chapter info is consistent!"
    )
    sess.execute(chapters.delete().where(chapters.c.course_id == basecourse))

    # chapters = Table("chapters", meta, autoload=True, autoload_with=engine)
    # sub_chapters = Table("sub_chapters", meta, autoload=True, autoload_with=engine)
    # questions = Table("questions", meta, autoload=True, autoload_with=engine)

    dynamic_pages = app.config.html_context.get("dynamic_pages", False)
    if dynamic_pages:
        cname = basecourse
    else:
        cname = app.env.config.html_context.get("course_id", "unknown")

    for chap in chaptitles:
        logger.info(f"Updating Database for {chap}")

        # insert row for chapter in the chapter table and get the id
        # logger.info("Adding Chapter/subchapter info for {}".format(chap))
        # check if chapter is already there
        sel = select([chapters]).where(
            and_(chapters.c.course_id == cname, chapters.c.chapter_label == chap)
        )
        res = sess.execute(sel).first()
        if not res:
            ins = chapters.insert().values(
                chapter_name=chaptitles.get(chap, chap),
                course_id=cname,
                chapter_label=chap,
                chapter_num=chap_numbers[chap],
            )
            res = sess.execute(ins)
            currentRowId = res.inserted_primary_key[0]
        else:
            currentRowId = res.id

        # If this chapter doesn't have subchapters, then skip them -- hence, use ``subtitles.get(chap, [])`` instead of ``subtitles[chap]``.
        for sub in subtitles.get(chap, []):
            if (chap, sub) in skips:
                skipreading = "T"
            else:
                skipreading = "F"
            # insert row for subchapter
            # todo: check if this chapter/subchapter is in the non-reading list
            q_name = "{}/{}".format(chaptitles.get(chap, chap), subtitles[chap][sub])
            ins = sub_chapters.insert().values(
                sub_chapter_name=subtitles[chap][sub],
                chapter_id=str(currentRowId),
                sub_chapter_label=sub,
                skipreading=skipreading,
                sub_chapter_num=subchap_numbers[chap][sub],
            )
            sess.execute(ins)
            # Three possibilities:
            # 1) The chapter and subchapter labels match existing, but the q_name doesn't match; because you changed
            # heading in a file.
            # 2) The chapter and subchapter labels don't match (new file name), but there is an existing q_name match,
            # because you renamed the file
            # 3) Neither match, so insert a new question
            sel = select([questions]).where(
                or_(
                    and_(
                        questions.c.chapter == chap,
                        questions.c.subchapter == sub,
                        questions.c.question_type == "page",
                        questions.c.base_course == basecourse,
                    ),
                    and_(
                        questions.c.name == q_name,
                        questions.c.question_type == "page",
                        questions.c.base_course == basecourse,
                    ),
                )
            )
            res = sess.execute(sel).first()
            if res and (
                (res.name != q_name) or (res.chapter != chap) or (res.subchapter != sub)
            ):
                # Something changed
                upd = (
                    questions.update()
                    .where(questions.c.id == res["id"])
                    .values(name=q_name, chapter=chap, from_source="T", subchapter=sub)
                )
                sess.execute(upd)
            if not res:
                # this is a new subchapter
                ins = questions.insert().values(
                    chapter=chap,
                    subchapter=sub,
                    question_type="page",
                    from_source="T",
                    name=q_name,
                    timestamp=datetime.now(),
                    base_course=basecourse,
                )
                sess.execute(ins)



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__author__ = "isaiahmayerchak"

from asyncore import write
from docutils import nodes
from docutils.parsers.rst import directives
from runestone.common.runestonedirective import (
    RunestoneIdDirective,
    RunestoneIdNode,
)
from runestone.server.componentdb import addQuestionToDB, addHTMLToDB
from runestone.common.xmlcommon import (
    write_substitute,
    substitute_departure,
)


def setup(app):
    app.add_directive("poll", Poll)
    app.add_node(
        PollNode,
        html=(visit_poll_html, depart_poll_html),
        xml=(visit_poll_xml, substitute_departure),
    )

    app.add_config_value("poll_div_class", "", "html")


TEMPLATE_START = """
<div class="runestone %(optclass)s">
<ul data-component="poll" id="%(divid)s" %(comment)s class='%(divclass)s' data-results='%(results)s' data-question_label="%(question_label)s" %(optional)s>
%(question)s
"""

TEMPLATE_OPTION = """
<li>%(optiontext)s</li>
"""

TEMPLATE_END = """</ul></div>"""


class PollNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


# self for these functions is an instance of the writer class.  For example
# in html, self is sphinx.writers.html.SmartyPantsHTMLTranslator
# The node that is passed as a parameter is an instance of our node class.
def visit_poll_html(self, node):
    res = visit_poll_common(self, node)
    addHTMLToDB(
        node["runestone_options"]["divid"], node["runestone_options"]["basecourse"], res
    )
    self.body.append(res)


def visit_poll_xml(self, node):
    html = visit_poll_common(self, node)
    write_substitute(self, node, html)


def depart_poll_html(self, node):
    """This is called at the start of processing a poll node.  If poll had recursive nodes
    etc and did not want to do all of the processing in visit_poll_html any finishing touches could be
    added here.
    """
    pass


def visit_poll_common(self, node):
    res = TEMPLATE_START
    res = res % node["runestone_options"]

    if node["runestone_options"]["scale"] == "":
        okeys = list(node["runestone_options"].keys())
        okeys.sort()
        i = 1
        for k in okeys:
            if "option_" in k:
                node["runestone_options"]["optiontext"] = (
                    f"{i}. " + node["runestone_options"][k]
                )
                i += 1
                res += TEMPLATE_OPTION % node["runestone_options"]
    else:
        for i in range(node["runestone_options"]["scale"]):
            node["runestone_options"]["optiontext"] = i + 1
            res += TEMPLATE_OPTION % node["runestone_options"]
    res += TEMPLATE_END

    return res


class Poll(RunestoneIdDirective):
    """
    .. poll:: identifier
        :scale: <X>  Setting the scale creates an "On a scale of 1 to <X>" type of question
        :allowcomment: Boolean--provides comment box
        :option_1: Providing Question text for each option creates a "Choose one of these options" type of poll.
        :option_2: Option 2
        :option_3: Option 3    ...etc...(Up to 10 options in mode 2)
        :results: One of all, instructor, superuser - who should see results?


    config values (conf.py):

    - poll_div_class - custom CSS class of the component's outermost div
    """

    required_arguments = 1
    optional_arguments = 0
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "scale": directives.positive_int,
            "allowcomment": directives.flag,
            "option_1": directives.unchanged,
            "option_2": directives.unchanged,
            "option_3": directives.unchanged,
            "option_4": directives.unchanged,
            "option_5": directives.unchanged,
            "option_6": directives.unchanged,
            "option_7": directives.unchanged,
            "option_8": directives.unchanged,
            "option_9": directives.unchanged,
            "option_10": directives.unchanged,
            "results": directives.unchanged,
        }
    )

    def run(self):
        """
        process the multiplechoice directive and generate html for output.
        :param self:
        :return:
        .. poll:: id
            :scale: <X>  Setting the scale creates an "On a scale of 1 to <X>" type of question
            :allowcomment: Boolean--provides comment box
            :option_1: Providing Question text for each option creates a "Choose one of these options" type of poll.
            :option_2: Option 2
            :option_3: Option 3    ...etc...(Up to 10 options in mode 2)
            :results: One of all, instructor, superuser - who should see results?
        """
        super(Poll, self).run()
        addQuestionToDB(self)

        if self.content:
            source = "\n".join(self.content)
        else:
            source = "\n"
        self.options["question"] = source

        if not "scale" in self.options:
            self.options["scale"] = ""
        if "allowcomment" in self.options:
            self.options["comment"] = "data-comment"
        else:
            self.options["comment"] = ""

        if not "results" in self.options:
            self.options["results"] = "instructor"

        env = self.state.document.settings.env
        self.options["divclass"] = env.config.poll_div_class

        poll_node = PollNode()
        poll_node["runestone_options"] = self.options
        poll_node["source"], poll_node["line"] = self.state_machine.get_source_and_line(
            self.lineno
        )
        return [poll_node]



# *******************************
# |docname| - Timed Assessments
# *******************************
# Group together several exercises into an assessment
# Really we should treat this as a kind of assignment.
# But it has to be done indirectly, especially in the case of a selectquestion
# see `runestone/selectquestion/toctree`
# 1. When processing a timed assessment add an assignment to the database for the basecourse
# 2. Before processing the body of the assessment we can set a flag in the environment
#    so that the children will know they are part of an assignment.
# 3. During recursive processing questions should add themselves to the assignment.
#    ``selectquestions`` shoud add themselves as selectquestions so the assignment
#    ends up with the correct fixed number of questions.  The resolution of these
#    will need to be handled by the grader...  The simplest thing may be to send the
#    log the results under the id of the select question rather than the selected question
#    as this will allow for the analysis by competency area.
#

# License
# -------
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "isaiahmayerchak"

# Imports
# -------
from docutils import nodes
from docutils.parsers.rst import directives
from runestone.common.runestonedirective import (
    RunestoneIdDirective,
    RunestoneIdNode,
)
from runestone.server.componentdb import addAssignmentToDB

# Timed Assessment Implementation
# -------------------------------
# Everydirective uses setup to add itself to the applications and add any nodes


def setup(app):
    app.add_directive("timed", TimedDirective)
    app.add_node(TimedNode, html=(visit_timed_html, depart_timed_html))


class TimedNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


def visit_timed_html(self, node):
    # Set options and format templates accordingly

    if "timelimit" not in node["runestone_options"]:
        node["runestone_options"]["timelimit"] = ""
    else:
        node["runestone_options"]["timelimit"] = "data-time=" + str(
            node["runestone_options"]["timelimit"]
        )

    if "noresult" in node["runestone_options"]:
        node["runestone_options"]["noresult"] = "data-no-result"
    else:
        node["runestone_options"]["noresult"] = ""

    if "timedfeedback" in node["runestone_options"]:
        node["runestone_options"]["timedfeedback"] = "data-timedfeedback=true"
    else:
        node["runestone_options"]["timedfeedback"] = ""

    if "notimer" in node["runestone_options"]:
        node["runestone_options"]["notimer"] = "data-no-timer"
    else:
        node["runestone_options"]["notimer"] = ""

    if "nofeedback" in node["runestone_options"]:
        node["runestone_options"]["nofeedback"] = "data-no-feedback"
    else:
        node["runestone_options"]["nofeedback"] = ""

    if "fullwidth" in node["runestone_options"]:
        node["runestone_options"]["fullwidth"] = "data-fullwidth"
    else:
        node["runestone_options"]["fullwidth"] = ""

    if "nopause" in node["runestone_options"]:
        node["runestone_options"]["nopause"] = "data-no-pause"
    else:
        node["runestone_options"]["nopause"] = ""

    res = TEMPLATE_START % node["runestone_options"]
    self.body.append(res)


def depart_timed_html(self, node):
    # Set options and format templates accordingly
    res = TEMPLATE_END % node["runestone_options"]

    self.body.append(res)


# Templates to be formatted by node options
TEMPLATE_START = """
    <div class="timedAssessment %(optclass)s">
    <ul data-component="timedAssessment" data-question_label="%(question_label)s" %(timelimit)s id="%(divid)s" %(noresult)s %(nofeedback)s %(timedfeedback)s %(notimer)s %(fullwidth)s %(nopause)s>
    """

TEMPLATE_END = """</ul>
    </div>
    """


class TimedDirective(RunestoneIdDirective):
    """
    .. timed:: identifier
        :timelimit: Number of minutes student has to take the timed assessment--if not provided, no time limit
        :noresult: Boolean, doesn't display score
        :timedfeedback: Boolean, Show feedback even in timed mode
        :notimer: Boolean, doesn't show timer
        :nopause: Boolean do not show a pause button
        :fullwidth: Boolean, allows the items in the timed assessment to take the full width of the screen...

    """

    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "timelimit": directives.positive_int,
            "noresult": directives.flag,
            "timedfeedback": directives.flag,
            "nofeedback": directives.flag,  # backward compatibility
            "fullwidth": directives.flag,
            "notimer": directives.flag,
            "nopause": directives.flag,
        }
    )

    def run(self):
        """
        process the timed directive and generate html for output.
        :param self:
        :return:
        .. timed:: identifier
            :timelimit: Number of minutes student has to take the timed assessment--if not provided, no time limit
            :noresult: Boolean, doesn't display score
            :timedfeedback: Boolean, show feedback
            :notimer: Boolean, doesn't show timer
            :fullwidth: Boolean, allows the items in the timed assessment to take the full width of the screen
        ...
        """
        super(TimedDirective, self).run()
        self.assert_has_content()  # make sure timed has something in it

        if "timelimit" in self.options:
            timelimit = self.options["timelimit"]
        else:
            timelimit = None

        timed_node = TimedNode()
        timed_node["runestone_options"] = self.options
        (
            timed_node["source"],
            timed_node["line"],
        ) = self.state_machine.get_source_and_line(self.lineno)
        # Use the environment so that any parsed directives will know they
        # are inside a timed exam.
        env = self.state.document.settings.env
        name = self.arguments[0].strip()
        if not getattr(env, "in_timed", False):
            setattr(env, "in_timed", name)
        addAssignmentToDB(name, self.basecourse, is_timed="T", time_limit=timelimit)

        self.state.nested_parse(self.content, self.content_offset, timed_node)
        delattr(env, "in_timed")

        return [timed_node]



from enum import Enum

class QuestionType(Enum):
    ACTIVECODE = ("activecode", "Active Code")
    MCHOICE = ("mchoice", "Multiple Choice")
    SHORTANSWER = ("shortanswer", "Short Answer")
    VIDEO = ("video", "Video")
    CODELENS = ("codelens", "Codelens")
    YOUTUBE = ("youtube", "Youtube")
    SELECTQUESTION = ("selectquestion", "Select question")
    CLICKABLEAREA = ("clickablearea", "Clickable Area")
    DRAGNDROP = ("dragndrop", "Drag and Drop")
    FILLINTHEBLANK = ("fillintheblank", "Fill in the blank")
    PARSONSPROB = ("parsonsprob", "Parsonsprob")
    POLL = ("poll", "Poll")
    HPARSONS = ("hparsons", "HParsons")
    ACTEX = ("actex", "Actex")

    def to_dict(self):
        return {
            "value": self.value[0],
            "label": self.value[1],
        }

    def value_only(self):
        return self.value[0]

    @classmethod
    def all_types(cls):
        return [member for member in cls]



# Copyright (C) 2017 Wayne Brown
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst import Directive
import string
import re
import os

__author__ = "wayne brown"


# --------------------------------------------------------------------------
def setup(app):
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Define a webgldemo directive
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    app.add_directive("webgldemo", WebglDemo)

    app.add_autoversioned_stylesheet("webgldemo.css")

    app.add_node(WebglDemoNode, html=(visit_webgldemo_html, depart_webgldemo_html))

    app.connect("doctree-resolved", process_webgldemo_nodes)
    app.connect("env-purge-doc", purge_webgldemos)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Define a webglinteractive directive
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    app.add_directive("webglinteractive", WebglInteractive)

    app.add_autoversioned_stylesheet("webglinteractive.css")

    # CodeMirror syntax highlighting for various types of code

    app.add_autoversioned_javascript("webglinteractive.js", defer="")

    # Javascript for saving files to the client's hard drive
    app.add_autoversioned_javascript("FileSaver.min.js", defer="")

    app.add_node(
        WebglInteractiveNode,
        html=(visit_webglinteractive_html, depart_webglinteractive_html),
    )

    app.connect("doctree-resolved", process_webglinteractive_nodes)
    app.connect("env-purge-doc", purge_webglinteractive)


def find_relative_path(env, options):
    # Get the path to the current lesson being processed.
    documentName = env.temp_data["docname"]
    documentPath, documentName = os.path.split(documentName)

    # Determine how deep this lesson is nested into the textbook folders.
    path_list = documentPath.split(os.sep)
    # If there is no path, it still returns one empty string
    if len(path_list) > 1:
        levels_deep = len(path_list)
    else:
        if len(path_list[0]) > 0:
            levels_deep = 1
        else:
            levels_deep = 0

    # For a WebGL program that is in a sub-folder of the "_static" folders:
    # "../" gets to the "_static" folder
    # If levels_deep is 0, then "../" must change to "_static/"
    # If levels_deep is 1, then "../" must change to "../_static"
    # if levels_deep is 2, then "../" must change to "../../_static"
    # etc.

    folder_prefix = "_static" + os.sep
    for j in range(levels_deep):
        folder_prefix = ".." + os.sep + folder_prefix

    options["lib_folder_prefix"] = folder_prefix

    # "./" gets to the WebGL program folder
    # If levels_deep is 0, then "./" must change to ""
    # If levels_deep is 1, then "./" must change to "../"
    # if levels_deep is 2, then "./" must change to "../../"
    # etc.

    folder_prefix = ""
    for j in range(levels_deep):
        folder_prefix = ".." + os.sep + folder_prefix

    options["program_folder_prefix"] = folder_prefix


# ==========================================================================
class WebglDemo(Directive):
    """
.. webgldemo:: uniqueid
    :htmlprogram: path/to/program/webgl_program.html
    :width: -- width of rendering canvas in pixels (default is value set in HTML)
    :height: -- height of rendering canvas in pixels (default is value set in HTML)
    :width2: -- width of secondary rendering canvas in pixels (default is value set in HTML)
    :height2: -- height of secondary rendering canvas in pixels (default is value set in HTML)

    The following assumptions are made about the structure of a textbook that
    uses this directive:
        * All WebGL programs are stored under the _static folder, each in a
          separate folder. (This causes the files to be automatically copied
          to the "build" folder when a "runestone build" is performed.)
          Therefore, the path to a WebGL program will always be some like:
          _static/example_folder/example.html
        * The parent folder of all WebGL HTML programs will always be
          "_static". Therefore, ".." always refers to the "_static" folder.
        * When a WebGL program is included into a lesson web page, the
          references to other files must be adjusted based on the path
          from the lesson page to the "_static" folder.
        * A specific lesson can contain one or more webgl programs.
            * HTML ids for tags must be unique for each separate program.
            * Each webgldemo must have a unique id, which is used to rename
              HTML element tags and Javascript variables.
            * Your HTML program should name id's and variables with a "my_" prefix.
              When the HTML program is included in the web page, the webgldemo
              unique id is put in every location of "my_". This allows a demo
              program to be used more than once on a page. For example,
              if you have a webgldemo named W1 and you have tags in your
              HTML like my_canvas, my_button, etc., then those names are
              converted to W1_canvas, W1_button, etc. when the program is
              loaded into the web page.
            * A webgl program embedded in a lesson page always includes a
              link to open the example program in a separate window. This
              allows a student to use a browser's "view page source"
              functionality to view only the webgl program, separate from
              the textbook's lesson page.

    Because of the above assumptions and conventions, the HTML description of
    a webgl program is written as if it is a single, standalone program that
    can be loaded into the student's web browser from a sub-folder of the _static
    folder of the textbook.

    When a webgl directive is executed, it does the following:
        * It extracts the body of the webgl HTML code and inserts the contents
          into the textbook page being built.
        * It modifies the file paths to the Javascript files, shader files,
            and model files so that they can be loaded from the lesson page's
            location, not the location of the program in the _static folder.
            These changes are summarized as follows:
            Changes:  "../learn_webgl"  to  "../_static/learn_webgl"
            Changes:  "../shaders"      to  "../_static/shaders"
            Changes:  "../models"       to  "../_static/models"
            Changes:  "src="./"         to  "src="../' + node["webgl_components"]['htmlprogrampath'] + "/")
        * The Javascript code at the bottom of a webgl program needs to be
          in the HTML code so that the runestone build process can modify
          the paths to these webgl program resources. The "Learn Computer Graphics
          using Webgl" textbook explains that in a "normal" webgl program, the
          Javascript code at the bottom would typically be moved into an
          appropriate Javascript code file.

    The default size for a webgl canvas is 400 by 400 pixels. The size of the
    canvas can be modified using the optional directive parameters.

    A typical webglcode directive will look like this:

        .. webgldemo:: W1
            :htmlprogram: _static/01_example01/scale_about_origin.html
            :width: 200
            :height: 200

    """

    required_arguments = 1
    optional_arguments = 0
    has_content = False
    option_spec = {
        # HTML description of the canvas and controls. This is a full
        # HTML page description. Only the body of the HTML page is
        # placed in the the webgldemo display area. This file is saved
        # in the _static hierarchy of the web site, so the path to
        # the static folder must be pretended to the name.
        "htmlprogram": directives.unchanged,
        "width": directives.unchanged,
        "height": directives.unchanged,
        "width2": directives.unchanged,
        "height2": directives.unchanged,
    }

    def run(self):

        self.options["divid"] = self.arguments[0]

        if "htmlprogram" in self.options:
            path, name = os.path.split(self.options["htmlprogram"])

            self.options["htmlprogrampath"] = path  # .encode('ascii', 'ignore')
            self.options["htmlprogram"] = name  # .encode('ascii', 'ignore')

            # Determine how much the path to resources must be adjusted.
            find_relative_path(self.state.document.settings.env, self.options)

            if "width" in self.options:
                self.options["width"] = int(self.options["width"])
            else:
                self.options["width"] = 0

            if "height" in self.options:
                self.options["height"] = int(self.options["height"])
            else:
                self.options["height"] = 0

            if "width2" in self.options:
                self.options["width2"] = int(self.options["width2"])
            else:
                self.options["width2"] = 0

            if "height2" in self.options:
                self.options["height2"] = int(self.options["height2"])
            else:
                self.options["height2"] = 0
        else:
            print("ERROR: An htmlprogram value is required for a webgldemo directive.")

        # print("self.options = ", self.options)

        return [WebglDemoNode(self.options)]


# ==========================================================================
class WebglDemoNode(nodes.General, nodes.Element):
    def __init__(self, content):
        """
        Arguments:
        - `self`:
        - `content`:
        """
        super(WebglDemoNode, self).__init__()
        self.webgl_components = content


# This creates HTML elements for a webgldemo program to execute.
# The elements include a webgldemo_canvas and some HTML
# interactive controls.
#
# -----------------------------------------------------------------
# |  css: webgldemo_canvas                                        |
# |  canvas                                                       |
# -----------------------------------------------------------------
# |  css: webgldemo_controls                                      |
# |  <input> controls                                             |
# |                                                               |
# -----------------------------------------------------------------
#
# HTML and CSS objects for the above design:
# <div class="webgldemo_container">
#   all of the HTML code here comes from the body of the htmlprogram
#   <canvas class="webgldemo_canvas"> .. <br />
#   <div class="webgldemo_controls"> .. </div>
# </div>
#

# self for these functions is an instance of the writer class.  For example
# in html, self is sphinx.writers.html.SmartyPantsHTMLTranslator
# The node that is passed as a parameter is an instance of our node class.
def visit_webgldemo_html(self, node):
    # print("In visit_webgldemo_html, node["webgl_components"] = ", node["webgl_components"])

    # start of HTML
    res = "<!-- webgldemo start -->\n"
    res += "<div id='%(divid)s_webgldemo' class='webgldemo_container'>\n"

    # The HTML comes from the body of the HTMLprogram file
    # Read the HTML file
    filename = os.path.join(
        node["webgl_components"]["htmlprogrampath"], node["webgl_components"]["htmlprogram"]
    )
    file = open(filename, "r")
    fileContents = file.read()
    file.close()

    # Extract only the body of the HTML code
    startPos = fileContents.find("<body")
    startPos = fileContents.find(">", startPos) + 1
    endPos = fileContents.find("</body>")
    bodyCode = fileContents[startPos:endPos]

    # Update the paths to library files
    bodyCode = bodyCode.replace("../", node["webgl_components"]["lib_folder_prefix"])
    bodyCode = bodyCode.replace(
        'src="./',
        'src="'
        + node["webgl_components"]["program_folder_prefix"]
        + node["webgl_components"]["htmlprogrampath"]
        + os.sep,
    )

    # make the canvas a specific size
    if node["webgl_components"]["width"] > 0 and node["webgl_components"]["height"] > 0:
        width_re = re.compile(
            r'<canvas id="my_canvas" class="webgldemo_canvas" width="(\d+)" height="(\d+)">'
        )
        if width_re.search(bodyCode):
            new_text = (
                r'<canvas id="my_canvas" class="webgldemo_canvas" width="'
                + str(node["webgl_components"]["width"])
                + '" height="'
                + str(node["webgl_components"]["height"])
                + r'">'
            )
            bodyCode = width_re.sub(new_text, bodyCode)

    if node["webgl_components"]["width2"] > 0 and node["webgl_components"]["height2"] > 0:
        # make the secondary canvas a specific size
        width_re = re.compile(
            r'<canvas id="my_canvas_b" class="webgldemo_canvas" width="(\d+)" height="(\d+)">'
        )
        if width_re.search(bodyCode):
            new_text = (
                r'<canvas id="my_canvas_b" class="webgldemo_canvas" width="'
                + str(node["webgl_components"]["width2"])
                + '" height="'
                + str(node["webgl_components"]["height2"])
                + r'">'
            )
            bodyCode = width_re.sub(new_text, bodyCode)

    # Make all the ids and variables unique for this embedded program.
    bodyCode = bodyCode.replace("my_", node["webgl_components"]["divid"] + "_")

    # Make the first parameter of the scene function be the ID of this webgldemo
    bodyCode = bodyCode.replace('"my",', '"' + node["webgl_components"]["divid"] + '",')

    # Add the HTML body code to the page in the "canvas" area.
    res += bodyCode

    # Add a link to open the webgldemo program in a separate browser tab.
    programPath = node["webgl_components"]["htmlprogrampath"].replace("_static/", "")
    myLink = os.path.join(
        node["webgl_components"]["lib_folder_prefix"],
        programPath,
        node["webgl_components"]["htmlprogram"],
    )
    res += (
        '<a href="'
        + myLink
        + '" target="_blank">Open this webgl demo program in a new tab or window</a>'
    )

    # End of HTML
    res += "</div>\n"
    res += "<!-- webgldemo end -->"

    # Replace all of the placeholders in the HTML with the strings stored
    # in the node["webgl_components"] dictionary.
    res = res % node["webgl_components"]

    self.body.append(res)


# --------------------------------------------------------------------------
def depart_webgldemo_html(self, node):
    """
    This is called at the start of processing an activecode node.  If activecode had recursive nodes
    etc and did not want to do all of the processing in visit_webgldemo_html any finishing touches could be
    added here.
    """
    pass


# --------------------------------------------------------------------------
def process_webgldemo_nodes(app, env, docname):
    pass


# --------------------------------------------------------------------------
def purge_webgldemos(app, env, docname):
    pass


# ==========================================================================
#
# webglinteractive directive
#
# ==========================================================================
#
# This creates an area for webgl code interactive experimentation.
# The area takes this basic format:
#
# -----------------------------------------------------------------
# |     css: webgl_cmds                                           |
# |     commands: reload, view options, download, download edited |
# -----------------------------------------------------------------
# |  css: webgl_row2                                              |
# | |-------------------------| |-------------------------------| |
# | | css: webgl_editors      | | css: webgl_canvas             | |
# | | tabbed textareas for    | | canvas for drawing            | |
# | | for code editing        | |                               | |
# | |                         | |                               | |
# | |                         | | HTML options for program      | |
# | |                         | | control, checkboxes, sliders, | |
# | |                         | | etc.                          | |
# | |-------------------------| |-------------------------------| |
# ----------------------------- -----------------------------------
# |  css: webgl_output                                            |
# |  text feedback                                                |
# |                                                               |
# -----------------------------------------------------------------
# Options:
# The code should be able to be initially hidden.
# The text feedback area should be able to be initially hidden.
#
# HTML and CSS objects for the above design:
# <div class="webgl_container">
#   <div class="webgl_cmds"> .. </div>
#   <div class="webgl_row2>
#      <div class="webgl_editors">
#         <div class="webgl_editors_tabs"> ... </div>
#         <div class="webgl_editors_contents"> ... </div>
#      </div>
#      <div class="webgl_canvas">  ... </div>
#   </div>
#   <div class="webgl_output">
#      checkbox options
#      <div class="webgl_output_text">
#          output
#      </div>
#   </div>
# </div>

# Organization of files that make up a WebGL 3D graphics display
# --------------------------------------------------------------
#
# A webGL program has these 5 parts:
#   1. HTML web page description (must include a canvas element)
#   2. Javascript program that initializes and "drives" the graphics
#   3. Webgl Vertex and Fragment shader programs
#   4. 3D models (optional if the models are generated procedurally in JavaScript code)
#
# Some of these components will be used over and over again for different
# WebGL scenes. Other components will be unique to a specific 3D scene.
# Therefore, we need an organization that divides the files into two groups:
#   - the reusable components
#   - the unique components
#
# The reusable components are:
#    2. Javascript libraries
#    3. WebGL shaders
#    4. 3D models
# The unique components are :
#    1. the HTML page
#    2. the Javascript programs that initialize and drive the graphics,
#       and the unique Javascript programs for this scene.
#    3. the WebGL shaders.
#    4. 3D models that are unique to this scene.
#
# Conclusions:
#   - The only unique file will be the HTML file
#   - Javascript, shaders, and models will be in either location
#
# Assume the LearnWebGL textbook (ands perhaps other applications)
# is organized like this, where each chapter is in a separate folder.
#    Chapter_1 (folder)
#       Lesson (html)
#       Lesson (html)
#       Lesson (html)
#    Chapter_2 (folder)
#       Lesson (html)
#       Lesson (html)
#       Lesson (html)
#    Chapter_3 (folder)
#       Lesson (html)
#       Lesson (html)
#       Lesson (html)
#    _static (folder)
#       learn_webgl (folder)
#           JavaScript files for webgl programs, *.js
#       models (folder)
#           model files, *.obj, *.mtl, etc.
#       shaders (folder)
#           shader programs, *.vert and *.frag
#       webgl_program_01 (folder)
#           JavaScript files specific to this program
#       webgl_program_02 (folder)
#           JavaScript files specific to this program
#       etc.
#
# Note that the webgl program folders are stored in the _static folder
# so that they are automatically copied to the build folder when a
# runestone book is built.
#
# So now we can write the file paths to various components, where "pathToPage"
# is the path to the current web page:

# The reusable components are:
#    2. Javascript libraries ( ../_static/learn_webgl/filename.js )
#    3. WebGL shaders        ( ../_static/shaders/filename.vert, *.frag )
#    4. 3D models            ( ../_static/models/model.obj, *.mtl )
# The unique components are :
#    1. the HTML page        ( ./filename.html )
#    2. Javascript programs  ( ./filename.js )
#    3. WebGL shaders.       ( ./filename.vert, ./filename.frag )
#    4. 3D models.           ( ./filename.obj, ./filename.mtl )
#

# =====================================================================
class WebglInteractive(Directive):
    """
.. webglinteractive:: uniqueid
    :htmlprogram: path/to/program/webgl_program.html
    :editList: comma separated list of files that can be edited
    :viewList: comma sepatated list of files that can be viewed, but not edited
    :width: -- width of rendering canvas in pixels (default is value in HTML)
    :height: -- height of rendering canvas in pixels (default is value in HTML)
    :width2: -- width of secondary rendering canvas in pixels (default is value in HTML)
    :height2: -- height of secondary rendering canvas in pixels (default is value in HTML)
    :hidecode: -- if present, the code is not initially displayed
    :hideoutput: -- if present, the output area below the code is not initially displayed

    The following assumptions are made about the structure of a textbook that
    uses this directive:
        * A lesson can contain one or more webgl programs.
            * HTML ids for tags must be unique for each separate program.
            * Each webgldemo must have a unique id, which is used to rename
              HTML element tags and Javascript variables.
            * Your HTML program should name id's and variables with a "my_" prefix.
              When the HTML program is included in the web page, the webgldemo
              unique id is put in every location of "my_". This allows a demo
              program to be used more than once on a page. For example,
              if you have a webgldemo named W1 and you have tags in your
              HTML like my_canvas, my_button, etc., then those names are
              converted to W1_canvas, W1_button, etc. when the program is
              loaded into the web page.
            * A webgl program embedded in a lesson page always includes a
              link to open the example program in a separate window. This
              allows a student to use a browser's "view page source"
              functionality to view only the webgl program, separate from
              the textbook's lesson page.
        * The example webgl programs included in a lesson are stored in the
          _static folders. (All files in the _static folders are automatically
          copied to the _static folder of the deployed textbook.)
        * Each webgl program has its own folder.

    Because of the above assumptions and conventions, the HTML description of
    a webgl program is written as if it is a single, standalone program that
    can be loaded into the student's web browser from a sub-folder of the _static
    folder of the textbook.

    When a webgl directive is executed, it does the following:
        * It extracts the body of the webgl HTML code and inserts the contents
          into the textbook page being built.
        * It modifies the file paths to the Javascript files, shader files,
            and model files so that they can be loaded from the lesson page's
            location, not the location of the program in the _static folder.
            These changes are summarized as follows:
            Changes:  "../learn_webgl"  to  "../_static/learn_webgl"
            Changes:  "../shaders"      to  "../_static/shaders"
            Changes:  "../models"       to  "../_static/models"
            Changes:  "src="./"         to  "src="../' + node["webgl_components"]['htmlprogrampath'] + "/")
        * The Javascript code at the bottom of a webgl program needs to be
          in the HTML code so that the runestone build process can modify
          the paths to these webgl program resources. The "Learn Computer Graphics
          usign Webgl" textbook explains that in a "normal" webgl program, the
          Javascript code at the bottom would typically be moved into an
          appropriate Javascript code file.

    The default size for a webgl canvas is 400 by 400 pixels. The size of the
    canvas can be modified using the optional directive parameters.

    A typical webglinteactive directive will look like this:

        .. webglinteractive:: W1
            :htmlprogram: _static/01_example01/scale_about_origin.html
            :editlist: _static/01_example01/scale_about_origin_render.js, _static/01_example01/model.js
            :width: 200
            :height: 200

    """

    required_arguments = 1
    optional_arguments = 0
    has_content = False
    option_spec = {
        # HTML description of the canvas and controls. This is a full
        # HTML page description. Only the body of the HTML page is
        # placed in the the webgl display area. This files are saved
        # in the _static hierarchy of the web site, so the path to
        # the static hierarchy must be pretended to the name.
        "htmlprogram": directives.unchanged,
        # A list of source files that are displayed in editor panes. Again
        # the path to these files must be adjusted.
        "editlist": directives.unchanged,
        "viewlist": directives.unchanged,
        # Options for hiding parts of a webglinteractive
        "hidecode": directives.flag,
        "hideoutput": directives.flag,
        # Change the size of the canvas
        "width": directives.unchanged,
        "height": directives.unchanged,
        "width2": directives.unchanged,
        "height2": directives.unchanged,
    }

    def run(self):
        env = self.state.document.settings.env

        # For debugging - print all the environment values
        # print("env: ")
        # attrs = vars(env)
        # for item in attrs.items():
        #     print item

        if not hasattr(env, "webglinteractivecounter"):
            env.webglinteractivecounter = 0
        env.webglinteractivecounter += 1

        self.options["divid"] = self.arguments[0]

        self.options["hidecanvas"] = False
        self.options["hideoutput"] = "hideoutput" in self.options
        self.options["hidecode"] = "hidecode" in self.options

        if "htmlprogram" in self.options:
            htmlprogram_folders = self.options["htmlprogram"].split("/")
            htmlprogram_path = "/".join(htmlprogram_folders[0:-1])
            htmlprogram_path = htmlprogram_path.strip()
            self.options["htmlprogram"] = htmlprogram_folders[-1]
            self.options["htmlprogrampath"] = htmlprogram_path
        else:
            print(
                "***** ERROR: An htmlprogram is required for a webglinteractive directive."
            )

        if "editlist" in self.options:
            # Turn the comma separated string of filenames into a list
            self.options["editlist"] = self.options["editlist"].split(",")
        else:
            self.options["editlist"] = []

        if "viewlist" in self.options:
            # Turn the comma separated string of filenames into a list
            self.options["viewlist"] = self.options["viewlist"].split(",")
        else:
            self.options["viewlist"] = []

        if "width" in self.options:
            self.options["width"] = int(self.options["width"])
        else:
            self.options["width"] = 0

        if "height" in self.options:
            self.options["height"] = int(self.options["height"])
        else:
            self.options["height"] = 0

        if "width2" in self.options:
            self.options["width2"] = int(self.options["width2"])
        else:
            self.options["width2"] = 0

        if "height2" in self.options:
            self.options["height2"] = int(self.options["height2"])
        else:
            self.options["height2"] = 0

        # Determine how much the path to resources must be adjusted.
        find_relative_path(self.state.document.settings.env, self.options)

        # print("edit list: ", self.options['editlist'])
        # print("view list: ", self.options['viewlist'])
        # print("self.options = ", self.options)

        return [WebglInteractiveNode(self.options)]


OUTPUT_OPTIONS = """
<div>
Show:
<input type="checkbox" id="%(divid)s_webgl_displayInfo" name="Display" value="InfoMessages" checked='checked' />
<span id="webgl_infoMessages" class="webgl_infoMessages">Process information &nbsp&nbsp</span>
<input type="checkbox" id="%(divid)s_webgl_displayWarnings" name="Display" value="Warnings" checked='checked' />
<span id="webgl_warningMessages" class="webgl_warningMessages">Warnings &nbsp&nbsp</span>
<input type="checkbox" id="%(divid)s_webgl_displayErrors" name="Display" value="Errors" checked='checked' />
<span id="webgl_errorMessages" class="webgl_errorMessages">Errors</span>
</div>
"""

OUTPUT_WINDOW = """
<div id="%(divid)s_webgl_output_div" class="webgl_output_div">
</div>
"""

# For the tabbed content
# There is a list of tab headings
# Then a list of the contents for each tab.

TAB_BEGIN = """<div id='%(divid)s_editors' class='webgl_code'>"""
TAB_END = """</div>"""

# Run the show/hide function on each element to get it into the correct
# visibility and width
SHOW_HIDE_SCRIPT = """
    <script>
        %(divid)s_directive.show_webgl(\"%(divid)s_show_code\",1);
        %(divid)s_directive.show_webgl(\"%(divid)s_show_canvas\",2);
        %(divid)s_directive.show_webgl(\"%(divid)s_show_info\",3);
        %(divid)s_directive.set_height_of_webgl_editors(\"%(divid)s\");
        %(divid)s_directive.bring_first_editor_to_front(\"%(divid)s\");
    </script>
"""

TAB_TITLES_LIST_BEGIN = """<ul class='webgl_nav_tabs' id='%(divid)s_tab'>"""

TAB_TITLES_LIST_END = """</ul>"""

TAB_LIST_ELEMENT = """
<li>
    <a data-toggle='tab' href='#%(divid)s_%(tabname)s'><span>%(tabtitle)s</span></a>
</li>
"""

TAB_CONTENTS_BEGIN = """<div style='width:100%%; position:relative;'>
<div class="webgl_code">"""

TAB_CONTENTS_END = """
<div style="clear:both;"></div>
</div>
</div>
"""

TAB_DIV_BEGIN = """<div class='webgl_tab_content' id='%(divid)s_%(tabname)s'>"""

TAB_DIV_END = """</div>"""

TABBED_EDITOR = """
<textarea id="%(tabid)s_textarea" class="webgl_tabbed_editor">
%(fileContents)s
</textarea>

<script>
    %(divid)s_directive.createCodeMirrorEditor('%(tabid)s', '%(filename)s', '%(fileextension)s' %(readonly)s);
</script>
"""


# =====================================================================
class WebglInteractiveNode(nodes.General, nodes.Element):
    def __init__(self, content):
        """

        Arguments:
        - `self`:
        - `content`:
        """
        super(WebglInteractiveNode, self).__init__()
        self.webgl_components = content


def add_commands(options):
    res = ""
    res += "<button class='btn btn-success' id='%(divid)s_runb' onclick='%(divid)s_directive.restart();'>Re-start</button>\n"
    res += "Show: \n"

    if options["hidecode"]:
        res += "<input type='checkbox' id='%(divid)s_show_code' onclick='%(divid)s_directive.show_webgl(\"%(divid)s_show_code\",1);' />Code &nbsp\n"
    else:
        res += "<input type='checkbox' id='%(divid)s_show_code' onclick='%(divid)s_directive.show_webgl(\"%(divid)s_show_code\",1);' checked='checked' />Code &nbsp\n"

    if options["hidecanvas"]:
        res += "<input type='checkbox' id='%(divid)s_show_canvas' onclick='%(divid)s_directive.show_webgl(\"%(divid)s_show_canvas\",2);' />Canvas &nbsp\n"
    else:
        res += "<input type='checkbox' id='%(divid)s_show_canvas' onclick='%(divid)s_directive.show_webgl(\"%(divid)s_show_canvas\",2);' checked='checked' />Canvas &nbsp\n"

    if options["hideoutput"]:
        res += "<input type='checkbox' id='%(divid)s_show_info' onclick='%(divid)s_directive.show_webgl(\"%(divid)s_show_info\",3);'/>Run Info\n"
    else:
        res += "<input type='checkbox' id='%(divid)s_show_info' onclick='%(divid)s_directive.show_webgl(\"%(divid)s_show_info\",3);' checked='checked' />Run Info\n"

    res += "<button class='btn webgl-default' id='%(divid)s_saveall' onclick='%(divid)s_directive.downloadAllFiles();'\n"
    res += "title='Download a copy of all the original\nfiles required for this WebGL program.'>Download Files</button>\n"

    res += "<button class='btn webgl-default' id='%(divid)s_save_edited' onclick='%(divid)s_directive.downloadEditedFiles();'\n"
    res += "title='Download a copy of all the\nfiles in the editor panels.'>Download Edited Files</button>\n"

    return res


def change_html_code_ids(bodyCode, webgl_components):

    if webgl_components["width"] > 0 and webgl_components["height"] > 0:
        # make the canvas a specific size
        width_re = re.compile(
            r'<canvas id="my_canvas" class="webgldemo_canvas" width="(\d+)" height="(\d+)">'
        )
        if width_re.search(bodyCode):
            new_text = (
                r'<canvas id="my_canvas" class="webgldemo_canvas" width="'
                + str(webgl_components["width"])
                + '" height="'
                + str(webgl_components["height"])
                + r'">'
            )
            bodyCode = width_re.sub(new_text, bodyCode)

    if webgl_components["width2"] > 0 and webgl_components["height2"] > 0:
        # make the secondary canvas a specific size
        width_re = re.compile(
            r'<canvas id="my_canvas_b" class="webgldemo_canvas" width="(\d+)" height="(\d+)">'
        )
        if width_re.search(bodyCode):
            new_text = (
                r'<canvas id="my_canvas_b" class="webgldemo_canvas" width="'
                + str(webgl_components["width2"])
                + '" height="'
                + str(webgl_components["height2"])
                + r'">'
            )
            bodyCode = width_re.sub(new_text, bodyCode)

    # Make all the ids and variables unique for this embedded program.
    bodyCode = bodyCode.replace("my_", webgl_components["divid"] + "_")

    # Make the first parameter of the scene function be the ID of this webgldemo
    bodyCode = bodyCode.replace('"my",', '"' + webgl_components["divid"] + '",')

    return bodyCode


def add_code_editors(options):
    # Put the code in the text editors in tabs
    res = ""

    # Add the tab headings
    res += TAB_TITLES_LIST_BEGIN

    number_to_edit = len(options["editlist"])
    for j in range(0, number_to_edit):

        tab_name = options["editlist"][j]
        # Remove any path on the file name
        folders = tab_name.split("/")
        if len(folders) > 1:
            tab_title = folders[len(folders) - 1].strip()
        else:
            tab_title = tab_name.strip()

        # Replace any periods in the filename to create the HTML id for the tab
        tab_name = tab_title.replace(".", "_").strip()

        res += TAB_LIST_ELEMENT % {
            "divid": options["divid"],
            "tabname": tab_name,
            "tabtitle": tab_title,
        }

    number_to_view = len(options["viewlist"])
    for j in range(0, number_to_view):

        tab_name = options["viewlist"][j]
        # Remove any path on the file name
        folders = tab_name.split("/")
        if len(folders) > 1:
            tab_title = folders[len(folders) - 1].strip()
        else:
            tab_title = tab_name.strip()

        # Replace any periods in the filename to create the HTML id for the tab
        tab_name = tab_title.replace(".", "_").strip()

        res += TAB_LIST_ELEMENT % {
            "divid": options["divid"],
            "tabname": tab_name,
            "tabtitle": tab_title,
        }

    res += TAB_TITLES_LIST_END

    # Add the tab contents
    res += TAB_CONTENTS_BEGIN

    for j in range(0, number_to_edit):
        edit_name = options["editlist"][j]
        # Remove any path on the file name
        folders = edit_name.split("/")
        if len(folders) > 1:
            file_name = folders[len(folders) - 1].strip()
        else:
            file_name = edit_name.strip()

        # Get the file name extension
        dot_pos = file_name.find(".")
        fileExtension = file_name[dot_pos + 1 :]

        # Replace any periods in the filename to create the HTML id for the tab
        tab_name = file_name.replace(".", "_").strip()

        tab_id = options["divid"] + "_" + tab_name
        tab_file = options["editlist"][j].strip()
        tab_file = os.path.abspath(tab_file)

        # Read the data file into a single string
        file = open(tab_file, "r")
        fileContents = file.read()
        file.close()

        # If this file is an HTML code file, then replace the "my_" code strings
        # with the directive ID. E.g., "my_" becomes "W2_"
        if fileExtension == "html":
            fileContents = change_html_code_ids(fileContents, options)

        res += TAB_DIV_BEGIN % {"divid": options["divid"], "tabname": tab_name}
        # content of tab
        res += TABBED_EDITOR % {
            "divid": options["divid"],
            "tabid": tab_id,
            "tabfile": edit_name,
            "fileContents": fileContents,
            "filename": file_name,
            "fileextension": fileExtension,
            "readonly": "",
        }
        res += TAB_DIV_END

    for j in range(0, number_to_view):
        view_name = options["viewlist"][j]
        # Remove any path on the file name
        folders = view_name.split("/")
        if len(folders) > 1:
            file_name = folders[len(folders) - 1].strip()
        else:
            file_name = view_name.strip()

        # Get the file name extension
        dot_pos = file_name.find(".")
        fileExtension = file_name[dot_pos + 1 :]

        # Replace any periods in the filename to create the HTML id for the tab
        tab_name = file_name.replace(".", "_").strip()

        tab_id = options["divid"] + "_" + tab_name
        tab_file = options["viewlist"][j].strip()
        tab_file = os.path.abspath(tab_file)

        # Read the data file into a single string
        file = open(tab_file, "r")
        fileContents = file.read()
        file.close()

        if fileExtension == "html":
            fileContents = change_html_code_ids(fileContents, options)

        res += TAB_DIV_BEGIN % {"divid": options["divid"], "tabname": tab_name}
        # content of tab
        res += TABBED_EDITOR % {
            "divid": options["divid"],
            "tabid": tab_id,
            "tabfile": view_name,
            "fileContents": fileContents,
            "filename": file_name,
            "fileextension": fileExtension,
            "readonly": ", true",
        }
        res += TAB_DIV_END

    res += TAB_CONTENTS_END

    return res


# self for these functions is an instance of the writer class.  For example
# in html, self is sphinx.writers.html.SmartyPantsHTMLTranslator
# The node that is passed as a parameter is an instance of our node class.
def visit_webglinteractive_html(self, node):
    # print("In visit_webglinteractive_html, node["webgl_components"] = ", node["webgl_components"])

    # Overall structure is here. Content is added in the functions
    res = "<!-- webglinteractive start -->"

    # Create an instance of the WebglInteractive_directive class. This creates
    # the event handlers and functionality for the webglinteractive directive.
    res += """
  <script>
    var %(divid)s_directive = new WebglInteractive_directive( '%(divid)s' );
  </script>
  """

    res += "<div class='webgl_container' id='%(divid)s_webgl_container'>\n"

    # Commands (top row)
    res += "<div class='webgl_cmds' id='%(divid)s_webgl_cmds'>\n"
    res += add_commands(node["webgl_components"])
    res += "</div>\n"

    # Start first row (editors and canvas)
    res += "<div class='webgl_row2' id='%(divid)s_webgl_row2'>\n"

    # Editors for code
    res += "<div class='webgl_editors' id='%(divid)s_webgl_editors'>\n"
    res += add_code_editors(node["webgl_components"])
    res += "</div>\n"

    # Canvas for rendering (with controls below the canvas)
    res += "<div class='webgl_canvas' id='%(divid)s_webgl_canvas'>"
    # Read the HTML file
    filename = (
        node["webgl_components"]["htmlprogrampath"]
        + "/"
        + node["webgl_components"]["htmlprogram"]
    )
    filename = os.path.realpath(filename)

    file = open(filename, "r")
    fileContents = file.read()
    file.close()

    # Put the HTML source file name so it can be downloaded by the user
    html_file_name = (
        "../"
        + node["webgl_components"]["htmlprogrampath"]
        + "/"
        + node["webgl_components"]["htmlprogram"]
    )
    res += '<span style="display: none;">' + html_file_name + "</span>"

    # Extract only the body of the HTML code
    startPos = fileContents.find("<body")
    startPos = fileContents.find(">", startPos) + 1
    endPos = fileContents.find("</body>")
    bodyCode = fileContents[startPos:endPos]

    # Update the paths to library files
    bodyCode = bodyCode.replace("../", node["webgl_components"]["lib_folder_prefix"])
    bodyCode = bodyCode.replace(
        'src="./',
        'src="'
        + node["webgl_components"]["program_folder_prefix"]
        + node["webgl_components"]["htmlprogrampath"]
        + os.sep,
    )

    bodyCode = change_html_code_ids(bodyCode, node["webgl_components"])

    # Add the HTML body code to the page in the "canvas" area.
    res += bodyCode

    res += "</div>\n"  # end of webgl_canvas

    res += "<div style='clear:both;'></div>"  # turns off the float left
    res += "</div>\n"

    # Output div for text feedback
    res += "<div class='webgl_output' id='%(divid)s_webgl_output'>"
    res += OUTPUT_OPTIONS
    res += OUTPUT_WINDOW
    res += "</div>\n"
    res += "<div style='clear:both;'></div>\n"  # turns off the float left

    # Add a link to open the webgl program in a separate browser tab.
    programPath = node["webgl_components"]["htmlprogrampath"].replace("_static/", "")
    myLink = os.path.join(
        node["webgl_components"]["lib_folder_prefix"],
        programPath,
        node["webgl_components"]["htmlprogram"],
    )
    res += (
        '<a href="'
        + myLink
        + '" target="_blank">Open this webgl program in a new tab or window</a>'
    )

    res += "</div>\n"

    # Add javascript code
    res += SHOW_HIDE_SCRIPT

    res += "<!-- webglinteractive end -->"

    # Separate the webgl components from the next content
    res += "<p> </p>\n"

    # Replace all of the placeholders in the HTML with the strings stored
    # in the node["webgl_components"] dictionary.
    res = res % node["webgl_components"]

    self.body.append(res)


def depart_webglinteractive_html(self, node):
    """ This is called at the start of processing an activecode node.  If activecode had recursive nodes
        etc and did not want to do all of the processing in depart_webglinteractive_html any finishing touches could be
        added here.
    """
    pass


def process_webglinteractive_nodes(app, env, docname):
    pass


def purge_webglinteractive(app, env, docname):
    pass



# *********
# |docname|
# *********
# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "bmiller"

from docutils import nodes
from docutils.parsers.rst import directives
from sqlalchemy import Table
from runestone.server.componentdb import (
    addQuestionToDB,
    addHTMLToDB,
    maybeAddToAssignment,
)
from runestone.common.runestonedirective import (
    RunestoneIdDirective,
    RunestoneIdNode,
)


def setup(app):
    app.add_directive("hparsons", HParsonsDirective)
    app.add_node(HParsonsNode, html=(visit_hp_html, depart_hp_html))


TEMPLATE_START = """
<div class="runestone">
<div data-component="hparsons" id=%(divid)s data-question_label="%(question_label)s" class="alert alert-warning hparsons_section">
<div class="hp_question">
"""

TEMPLATE_END = """
</div>
<div class='hparsons'></div>
<textarea
    %(language)s
    %(optional)s
    %(dburl)s
    %(reuse)s
    %(randomize)s
    %(blockanswer)s
    style="visibility: hidden;">
%(initialsetting)s
</textarea>
</div>
</div>
"""


class HParsonsNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


# self for these functions is an instance of the writer class.  For example
# in html, self is sphinx.writers.html.SmartyPantsHTMLTranslator
# The node that is passed as a parameter is an instance of our node class.
def visit_hp_html(self, node):

    node["delimiter"] = "_start__{}_".format(node["runestone_options"]["divid"])

    self.body.append(node["delimiter"])

    res = TEMPLATE_START % node["runestone_options"]
    self.body.append(res)


def depart_hp_html(self, node):
    res = TEMPLATE_END % node["runestone_options"]
    self.body.append(res)

    addHTMLToDB(
        node["runestone_options"]["divid"],
        node["runestone_options"]["basecourse"],
        "".join(self.body[self.body.index(node["delimiter"]) + 1 :]),
    )

    self.body.remove(node["delimiter"])


class HParsonsDirective(RunestoneIdDirective):
    """
    .. hparsons:: uniqueid
       :language: python, java, javscript, sql, html: only for highlighting purpose.
       :dburl: only for sql -- url to load database
       :randomize: randomize the order of horizontal parsons
       :reuse: only for parsons -- make the blocks reusable
       :blockanswer: 0 1 2 3 # Provide answer for block-based feedback. Please note that the number of block start from 0. If not provided, will use execution based feedback.

        Here is the problem description. It must ends with the tildes.
        Make sure you use the correct delimitier for each section below.
        ~~~~
        --hiddenprefix--
        // code that is for scaffolding the execution (e.g. initializing database)
        --blocks--
        block 1
        block 2
        block 3
        --hiddensuffix--
        // code that is for scaffolding unittest/execution (e.g. adding query for database)
        // most of the time the hiddensuffix is just "select * from table" to
        //      get all entries from the table to test the update or other operations.
        --unittest--
        assert 1,1 == world
        assert 0,1 == hello
        assert 2,1 == 42
    """

    required_arguments = 1
    optional_arguments = 1
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "dburl": directives.unchanged,
            "language": directives.unchanged,
            "reuse": directives.flag,
            "randomize": directives.flag,
            "blockanswer": directives.unchanged,
        }
    )

    def run(self):
        super(HParsonsDirective, self).run()
        addQuestionToDB(self)

        env = self.state.document.settings.env

        if "language" in self.options:
            self.options["language"] = "data-language='{}'".format(
                self.options["language"]
            )
        else:
            self.options["language"] = ""

        if "reuse" in self.options:
            self.options["reuse"] = ' data-reuse="true"'
        else:
            self.options["reuse"] = ""

        if "randomize" in self.options:
            self.options["randomize"] = ' data-randomize="true"'
        else:
            self.options["randomize"] = ""

        if "blockanswer" in self.options:
            self.options["blockanswer"] = "data-blockanswer='{}'".format(
                self.options["blockanswer"]
            )
        else:
            self.options["blockanswer"] = ""

        explain_text = None
        if self.content:
            if "~~~~" in self.content:
                idx = self.content.index("~~~~")
                explain_text = self.content[:idx]
                self.content = self.content[idx + 1 :]
            source = "\n".join(self.content)
        else:
            source = "\n"

        self.explain_text = explain_text or ["Not an Exercise"]

        self.options["initialsetting"] = source

        # SQL Options
        if "dburl" in self.options:
            self.options["dburl"] = "data-dburl='{}'".format(self.options["dburl"])
        else:
            self.options["dburl"] = ""

        course_name = env.config.html_context["course_id"]
        divid = self.options["divid"]

        hpnode = HParsonsNode()
        hpnode["runestone_options"] = self.options
        hpnode["source"], hpnode["line"] = self.state_machine.get_source_and_line(
            self.lineno
        )
        self.add_name(hpnode)  # make this divid available as a target for :ref:

        maybeAddToAssignment(self)
        if explain_text:
            self.updateContent()
            self.state.nested_parse(explain_text, self.content_offset, hpnode)

        return [hpnode]



# Copyright (C) 2019  Bradley N. Miller,
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "bmiller"

import csv
import io
import os
import re
import urllib.parse
from docutils import nodes
from docutils.parsers.rst import directives
from runestone.common.runestonedirective import (
    RunestoneNode,
    RunestoneIdDirective,
    RunestoneIdNode,
    get_node_line,
)
from runestone.server.componentdb import addQuestionToDB, addHTMLToDB

#
# setup is called when extensions are loaded. This registers the new directive and
# logs any js or css files that should be loaded for this extension.
#


def setup(app):
    app.add_directive("spreadsheet", SpreadSheet)

    app.add_node(SpreadSheetNode, html=(visit_ss_html, depart_ss_html))


# When the directive is process we will create nodes in the document tree to account
# for what we need to see on the final page.  Although we only care to render interactive
# textbooks as HTML one could, render the nodes as LaTex or many other languages.
#
class SpreadSheetNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


#
# The spreadsheet class implements the directive.
# When the directive is processed the run method is called.
# This allows us to handle any arguments, and then create a node or nodes to insert into the
# document tree to be rendered when the tree is written.
#


class SpreadSheet(RunestoneIdDirective):
    """
    .. spreadsheet:: uniqueid
        :fromcsv: path/to/csv/file
        :colwidths: list of column widths
        :coltitles: list of column names
        :mindimensions: mincols, minrows  -- minDimensions:[10,5]

        A1,B1,C1,D1...
        A2,B2,C2,D2...
    """

    required_arguments = 1
    optional_arguments = 5
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "fromcsv": directives.unchanged,
            "colwidths": directives.unchanged,
            "coltitles": directives.unchanged,
            "mindimensions": directives.unchanged,
        }
    )

    def run(self):
        super(SpreadSheet, self).run()
        env = self.state.document.settings.env

        self.options["divid"] = self.arguments[0].strip()

        if "====" in self.content:
            idx = self.content.index("====")
            suffix = self.content[idx + 1 :]
            self.options["asserts"] = suffix
            self.options["autograde"] = 'data-autograde="true"'
        else:
            self.options["asserts"] = '""'
            self.options["autograde"] = ""

        if "fromcsv" in self.options:
            self.content = self.body_from_csv(env, self.options["fromcsv"])
        else:
            if self.content:
                self.content = self.body_to_csv(self.content)
            else:
                raise ValueError(
                    "You must specify either from csv or provide content in the body"
                )

        self.options["data"] = self.content

        if "coltitles" not in self.options:
            self.options["coltitles"] = ""
        else:
            self.options["coltitles"] = "data-coltitles=[{}]".format(
                ",".join(
                    [
                        urllib.parse.quote(x.strip(), safe="'\"")
                        for x in self.options["coltitles"].split(",")
                    ]
                )
            )

        if "mindimensions" not in self.options:
            self.options["mindimensions"] = ""
        else:
            self.options["mindimensions"] = "data-mindimensions=[{}]".format(
                ",".join([x.strip() for x in self.options["mindimensions"].split(",")])
            )

        if "colwidths" not in self.options:
            self.options["colwidths"] = ""
        else:
            self.options["colwidths"] = "data-colwidths=[{}]".format(
                ",".join([x.strip() for x in self.options["colwidths"].split(",")])
            )

        ssnode = SpreadSheetNode()
        ssnode["runestone_options"] = self.options
        ssnode["source"], ssnode["line"] = self.state_machine.get_source_and_line(
            self.lineno
        )
        self.add_name(ssnode)  # make this divid available as a target for :ref:

        return [ssnode]

    def body_to_csv(self, row_list):
        """
        Use the csv reader and writer functionality to better parse the body.

        1. Convert the contents to a StringIO object
        2. Then read and process using a csv reader
        3. Formulas with ,'s in them should be in double quotes, if there are "'s in the
           cell then they should be ""-ed
        """

        csvlist = []
        body_list = []
        for row in row_list:
            if re.match(r"^\s*====", row):
                break
            body_list.append(row)
        body_file = io.StringIO("\n".join(body_list))
        body_reader = csv.reader(body_file)
        for row in body_reader:
            ilist = []
            for item in row:
                item = item.strip()
                if item and item[0] == '"' and item[-1] == '"':
                    item = item[1:-1]
                if is_float(item):
                    ilist.append(as_int_or_float(item))
                elif item.startswith("="):
                    ilist.append("{}".format(item.upper()))
                else:
                    ilist.append("{}".format(item))
            csvlist.append(ilist)
        return csvlist

    def body_from_csv(self, env, csvfile):
        ffpath = os.path.dirname(self.srcpath)
        print(self.srcpath, os.getcwd())
        filename = os.path.join(env.srcdir, ffpath, csvfile)

        print("\n\nPATH=", self.srcpath)
        with open(filename, "r") as csv:
            content = csv.readlines()

        content = [line[:-1] for line in content]
        return self.body_to_csv(content)


def is_float(s):
    try:
        x = float(s)
        return True
    except:
        return False


def as_int_or_float(s):
    try:
        x = int(s)
        return x
    except:
        x = float(s)
        return x


TEMPLATE = """
<div id="{divid}" data-component="spreadsheet" class="runestone %(optclass)s" {autograde} {mindimensions} {colwidths} {coltitles} data-question_label="{question_label}">
    <div id="{divid}_sheet"></div>

    <script>
        {divid}_data = {data};
        {divid}_asserts = {asserts};
    </script>

</div>
"""


def visit_ss_html(self, node):
    res = TEMPLATE.format(**node["runestone_options"])
    self.body.append(res)


def depart_ss_html(self, node):
    pass



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "bmiller"

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst import Directive
from runestone.server.componentdb import addQuestionToDB, addHTMLToDB
from runestone.common.runestonedirective import (
    RunestoneIdDirective,
    RunestoneDirective,
    RunestoneIdNode,
)


def setup(app):
    app.add_directive("video", Video)
    app.add_directive("youtube", Youtube)
    app.add_directive("vimeo", Vimeo)
    app.add_node(
        VideoNode,
        html=(visit_video_html, depart_video_html),
        xml=(visit_video_xml, depart_video_xml),
    )


CODE = """\
<div id="%(divid)s" class="video_popup runestone %(optclass)s" >
<video %(controls)s %(preload)s %(loop)s poster="%(thumb)s">
    %(sources)s
    No supported video types
</video>
</div>
"""

POPUP = """\
<a id="%(divid)s_thumb" style='position:relative;'>
    <img src="%(thumb)s" />
    <div class='video-play-overlay'></div>
</a>
<script>
  window.addEventListener('load', (event) => {
    jQuery(function ($) {
       $('#%(divid)s_thumb').click(function (e) {
                $('#%(divid)s').modal();
                return false;
        });
    });
  });
</script>

"""

INLINE = """\
<script>
  window.addEventListener('load', (event) => {
    jQuery(function($) {
        var rb = new RunestoneBase();
        $('#%(divid)s_thumb').click(function(e) {
            $('#%(divid)s').show();
            $('#%(divid)s_thumb').hide();
            rb.logBookEvent({'event':'video','act':'play','div_id': '%(divid)s'});
            // Log the run event
        });
        $('#%(divid)s video').one("click", function(){
            this.play();
        });
        $('#%(divid)s video').one("play", function(){
            rb.logBookEvent({'event':'video','act':'play','div_id': '%(divid)s'});
        });
    });
  });
</script>
"""
SOURCE = """<source src="%s" type="video/%s"></source>"""


XML_OUTPUT = """
    <video xml:id="{divid}" youtube="{video_id}" width="{width}" />
    """


class VideoNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


def visit_video_html(self, node):
    html = node["template"] % node["runestone_options"]
    self.body.append(html)
    addHTMLToDB(
        node["runestone_options"]["divid"],
        node["runestone_options"]["basecourse"],
        html,
    )


def visit_video_xml(self, node):
    xml = XML_OUTPUT.format(**node["runestone_options"])
    self.output.append(xml)


def depart_video_xml(self, node):
    pass


def depart_video_html(self, node):
    pass


class Video(RunestoneIdDirective):
    """
    .. video:: id
       :controls:  Show the controls or not
       :loop: loop the video
       :thumb: url to thumbnail image
       :preload: set the video to preload in the bg

       url to video format 1
       url to video format 2
       ...
    """

    required_arguments = 1
    optional_arguments = 1
    final_argument_whitespace = True
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "controls": directives.flag,
            "loop": directives.flag,
            "thumb": directives.uri,
            "preload": directives.flag,
        }
    )

    def run(self):
        """
        process the video directive and generate html for output.
        :param self:
        :return:
        """
        super(Video, self).run()
        addQuestionToDB(self)

        mimeMap = {"mov": "mp4", "webm": "webm", "m4v": "m4v"}

        sources = [
            SOURCE % (directives.uri(line), mimeMap[line[line.rindex(".") + 1 :]])
            for line in self.content
        ]
        if "controls" in self.options:
            self.options["controls"] = "controls"
        if "loop" in self.options:
            self.options["loop"] = "loop"
        else:
            self.options["loop"] = ""

        if "preload" in self.options:
            self.options["preload"] = 'preload="auto"'
        else:
            self.options["preload"] = 'preload="none"'

        self.options["sources"] = "\n    ".join(sources)
        res = CODE % self.options
        if "popup" in self.options:
            res += POPUP % self.options
        else:
            res += INLINE % self.options

        addHTMLToDB(self.options["divid"], self.options["basecourse"], res)
        return [nodes.raw(self.block_text, res, format="html")]


"""
    ReST directive for embedding Youtube and Vimeo videos.

    There are two directives added: ``youtube`` and ``vimeo``. The only
    argument is the video id of the video to include.

    Both directives have four optional arguments: ``height``, ``width``
    and ``align``. Default height is 281 and default width is 500.
    The default transport is HTTP although HTTPS can be specified.

    Example::

        .. youtube:: anwy2MPT5RE
            :height: 315
            :width: 560
            :align: left
            :http: http

    :copyright: (c) 2012 by Danilo Bargen.
    :license: BSD 3-clause
"""


def align(argument):
    """Conversion function for the "align" option."""
    return directives.choice(argument, ("left", "center", "right"))


def httpOption(argument):
    """Conversion function for the "http" option."""
    return directives.choice(argument, ("http", "https"))


class IframeVideo(RunestoneIdDirective):
    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update(
        {
            "height": directives.nonnegative_int,
            "width": directives.nonnegative_int,
            "align": align,
            "http": httpOption,
            "divid": directives.unchanged,
        }
    )
    default_width = 500
    default_height = 281

    def run(self):
        super(IframeVideo, self).run()
        self.options["video_id"] = directives.uri(self.arguments[0])
        if not self.options.get("width"):
            self.options["width"] = self.default_width
        if not self.options.get("height"):
            self.options["height"] = self.default_height
        if not self.options.get("align"):
            self.options["align"] = "center"
        if not self.options.get("http"):
            self.options["http"] = "https"
        if not self.options.get("divid"):
            self.options["divid"] = self.arguments[0]

        # res = self.html % self.options
        # addHTMLToDB(self.options["divid"], self.options["basecourse"], res)
        raw_node = VideoNode()
        raw_node["runestone_options"] = self.options
        raw_node["template"] = self.html
        # nodes.raw(self.block_text, res, format="html")
        raw_node["source"], raw_node["line"] = self.state_machine.get_source_and_line(
            self.lineno
        )
        return [raw_node]


class Youtube(IframeVideo):
    """
    .. youtube:: YouTubeID
       :divid: the runestone id for this video
       :height: 315
       :width: 560
       :align: left
       :http: http
       :start: None
       :end: None
    """

    html = """
    <div class="runestone %(optclass)s" style="margin-left: auto; margin-right:auto">
        <div id="%(divid)s" data-component="youtube" class="align-%(align)s youtube-video"
            data-video-height="%(height)d" data-question_label="%(question_label)s"
            data-video-width="%(width)d" data-video-videoid="%(video_id)s"
            data-video-divid="%(divid)s" data-video-start="%(start)d"
            data-video-end="%(end)s" ></div>
    </div>
    """

    option_spec = IframeVideo.option_spec.copy()
    option_spec.update(
        {"start": directives.nonnegative_int, "end": directives.nonnegative_int}
    )

    def run(self):
        if not self.options.get("start"):
            self.options["start"] = 0

        if not self.options.get("end"):
            self.options["end"] = -1

        raw_node = super(Youtube, self).run()
        addQuestionToDB(self)
        return raw_node


class Vimeo(IframeVideo):
    """
    .. vimeo:: vimeoID
       :height: 315
       :width: 560
       :align: left
       :http: http
    """

    html = '<iframe src="%(http)s://player.vimeo.com/video/%(video_id)s" \
    width="%(width)u" height="%(height)u" frameborder="0" \
    webkitAllowFullScreen mozallowfullscreen allowFullScreen \
    class="align-%(align)s" seamless ></iframe>'


source = """\
This is some text.

.. video:: divid
   :controls:
   :thumb: _static/turtlestill.png
   :loop:

   http://knuth.luther.edu/~bmiller/foo.mov
   http://knuth.luther.edu/~bmiller/foo.webm

This is some more text.
"""

if __name__ == "__main__":
    from docutils.core import publish_parts

    directives.register_directive("video", Video)

    doc_parts = publish_parts(
        source,
        settings_overrides={"output_encoding": "utf8", "initial_header_level": 2},
        writer_name="html",
    )

    print(doc_parts["html_body"])



"""
**********************************************
|docname| - select an accessibility stylesheet
**********************************************
By setting the `accessibility_style` config value (in `conf.py <../common/project_template/conf.tmpl>` of
an interactive book project) you can select what accessibility stylesheet
you want to add (``normal``, ``light``, ``darkest``, or ``none``).

An accessibility stylesheet covers:

-   Change of nav bar to color code based on Users using either mouse or keyboard to navigate the menu
-   Adjusting bootstrap buttons to invert color on active and on focus for accessibility for users
-   Changing default bootstrap buttons to follow WCAG 2.0 guidelines

| ``acessibility.css`` reflects WCAG 2.0 AA compliance
| ``acessibilitydarkest.css`` reflects WCAG 2.0 AAA compliance
| ``accessibilitylight.css`` doesn't change bootstrap colors but adds inversion

Personally we prefered WCAG 2.0 AA compliance, so ``accessibility.css``
reflects ideal changes

.. figure:: https://i.imgur.com/jFzkI6g.jpg

    This shows the different CSS files from most compliance (left) to least compliance (right).
"""


def setup(app):
    app.add_config_value("accessibility_style", "normal", "html")

    # Since the ``init_values()`` method of the ``app.config`` object (Config class)
    # is not invoked yet (will be invoked after the setup of all extensions),
    # we need to access the raw config dictionary here.
    acc_style = app.config._raw_config.get("accessibility_style", "normal")

    if acc_style == "normal":
        app.add_autoversioned_stylesheet("accessibility.css")
    elif acc_style == "light":
        app.add_autoversioned_stylesheet("accessibilitylight.css")
    elif acc_style == "darkest":
        app.add_autoversioned_stylesheet("accessibilitydarkest.css")



# ********************************
# |docname| - LP module definition
# ********************************
from runestone.lp import inlinesyntaxhighlight, lp

# Register the for loop directive with Sphinx.
from runestone.lp import for_loop  # NOQA: F401


# Initialize both extensions.
def setup(app):
    inlinesyntaxhighlight.setup(app)
    return lp.setup(app)



# Copyright (C) 2017  Tyler Conzett
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "tconzett"

from asyncore import write
from docutils import nodes
from docutils.parsers.rst import directives
from runestone.common.xmlcommon import substitute_departure
from runestone.server.componentdb import addQuestionToDB, addHTMLToDB
from runestone.common.runestonedirective import (
    RunestoneIdDirective,
    RunestoneIdNode,
)
from runestone.common.xmlcommon import (
    substitute_departure,
    write_substitute,
)


def setup(app):
    app.add_directive("showeval", ShowEval)

    app.add_config_value("showeval_div_class", "runestone explainer", "html")
    app.add_node(
        ShowEvalNode,
        html=(visit_showeval_html, depart_showeval_html),
        xml=(visit_showeval_xml, substitute_departure),
    )


# Create visitors, so we can generate HTML after the doctree is resolve (where the question label is determined).
class ShowEvalNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


def visit_showeval_html(self, node):
    html = CODE % node["runestone_options"]
    self.body.append(html)
    addHTMLToDB(
        node["runestone_options"]["divid"],
        node["runestone_options"]["basecourse"],
        html,
    )


def visit_showeval_xml(self, node):
    html = CODE % node["runestone_options"]
    write_substitute(self, node, html)


def depart_showeval_html(self, node):
    pass


CODE = """
<div data-component="showeval" data-question_label="%(question_label)s" class="%(divclass)s" id="%(divid)s" data-tracemode="%(trace_mode)s" %(optional)s>
    <button class="btn btn-success" id="%(divid)s_nextStep">Next Step</button>
    <button class="btn btn-default" id ="%(divid)s_reset">Reset</button>
    <div class="evalCont" style="background-color: #FDFDFD;">%(preReqLines)s</div>
    <div class="evalCont"></div>
    <script>
    if (typeof window.raw_steps === "undefined") {
    window.raw_steps = {};
    }
    raw_steps["%(divid)s"] = %(steps)s;
    </script>
</div>
"""


class ShowEval(RunestoneIdDirective):
    """
    .. showeval:: unique_id_goes_here
       :trace_mode: boolean  <- Required option that enables 'Trace Mode'

       some code
       more code
       ~~~~
       more {{code}}{{what code becomes in step 1}}
       ##Optional comment for step 2
       more {{what code becomes in step 1}}{{what code becomes in step2}}
       as many steps as you want {{the first double braces}}{{animate into the second}} wherever.


    config values (conf.py):

    - showeval_div_class - custom CSS class of the component's outermost div
    """

    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = RunestoneIdDirective.option_spec.copy()
    option_spec.update({"trace_mode": directives.unchanged_required})

    def run(self):
        """
            All prerequisite information that should be displayed above the directive,
            such as variable declaration, are separated from the step strings by "-----".

            The step animations follow the "-----" and are written one per line. Use
            "{{" and "}}" braces to surround the part of the line that should be replaced,
            followed by the replacement text also in "{{" and "}}". If you would like to add
            a comment that will appear in a div beside the animation, denote that at the end
            of the step where you would like it to appear with "##".

        Example:

        .. showeval:: showEval_0
           :trace_mode: false

           eggs = ['dogs', 'cats', 'moose']
           ~~~~

           ''.join({{eggs}}{{['dogs', 'cats', 'moose']}}).upper().join(eggs)
           ##I want to put a comment here!
           {{''.join(['dogs', 'cats', 'moose'])}}{{'dogscatsmoose'}}.upper().join(eggs)
           {{'dogscatsmoose'.upper()}}{{'DOGSCATSMOOSE'}}.join(eggs)
           'DOGSCATSMOOSE'.join({{eggs}}{{['dogs', 'cats', 'moose']}})
           {{'DOGSCATSMOOSE'.join(['dogs', 'cats', 'moose'])}}{{'dogsDOGSCATSMOOSEcatsDOGSCATSMOOSEmoose'}}

        """

        # Raise an error if the directive does not have contents.
        super(ShowEval, self).run()
        addQuestionToDB(self)

        self.options["trace_mode"] = self.options["trace_mode"].lower()
        self.options["preReqLines"] = ""
        self.options["steps"] = []

        env = self.state.document.settings.env
        self.options["divclass"] = env.config.showeval_div_class

        is_dynamic = env.config.html_context.get("dynamic_pages", False)
        is_dynamic = True if is_dynamic == "True" else False
        step = False
        for line in self.content:
            if step == True:
                if line != "":
                    if is_dynamic:
                        esc_line = str(line).replace("{", r"\{")
                    else:
                        esc_line = str(line)
                    self.options["steps"].append(esc_line)
            elif "~~~~" in line:
                step = True
            else:
                self.options["preReqLines"] += line + "<br />\n"
        se_node = ShowEvalNode()
        se_node["runestone_options"] = self.options
        se_node["source"], se_node["line"] = self.state_machine.get_source_and_line(
            self.lineno
        )
        return [se_node]



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "isaiahmayerchak"

from docutils import nodes
from docutils.parsers.rst import directives

# We don't use a ``RunestoneIdNode``, because these imply a numbered question, not a GUI element. Likewise, only ``tabbed`` requires a divid; in addition, this divid doesn't need to be globally unique, since it won't be connected with a question in the database. Therefore, just use a RunestoneDirectory for these.
from runestone.common.runestonedirective import (
    RunestoneDirective,
    RunestoneNode,
)


def setup(app):
    # Add directives/javascript/css
    app.add_directive("tabbed", TabbedStuffDirective)
    app.add_directive("tab", TabDirective)

    app.add_node(TabNode, html=(visit_tab_html, depart_tab_html))
    app.add_node(
        TabbedStuffNode, html=(visit_tabbedstuff_html, depart_tabbedstuff_html)
    )

    app.add_config_value("tabbed_div_class", "alert alert-warning", "html")


# Templates to be formatted by node options
BEGIN = """<div id='%(divid)s' data-component="tabbedStuff" %(inactive)s class='%(divclass)s %(optclass)s'>"""

TABDIV_BEGIN = """<div data-component="tab" data-tabname="%(tabname)s" %(active)s>
"""

TABDIV_END = """</div>"""

END = """
    </div>

"""


class TabNode(nodes.General, nodes.Element):
    pass


def visit_tab_html(self, node):
    # Set options and format templates accordingly
    divid = node.parent["divid"]
    tabname = node["tabname"]

    if "active" in node["tabnode_options"]:
        node["tabnode_options"]["active"] = "data-active"
    else:
        node["tabnode_options"]["active"] = ""

    res = TABDIV_BEGIN % {
        "divid": divid,
        "tabname": tabname,
        "active": node["tabnode_options"]["active"],
    }
    self.body.append(res)


def depart_tab_html(self, node):
    # Set options and format templates accordingly
    self.body.append(TABDIV_END)


class TabbedStuffNode(nodes.General, nodes.Element, RunestoneNode):
    """A TabbedStuffNode contains one or more TabNodes"""

    pass


def visit_tabbedstuff_html(self, node):
    divid = node["divid"]
    if "inactive" in node["tabbed_stuff_options"]:
        node["tabbed_stuff_options"]["inactive"] = "data-inactive"
    else:
        node["tabbed_stuff_options"]["inactive"] = ""

    format_dict = {
        "divid": divid,
        "divclass": node["tabbed_stuff_options"]["divclass"],
        "inactive": node["tabbed_stuff_options"]["inactive"],
    }

    format_dict.update(node["tabbed_stuff_options"])

    res = BEGIN % format_dict

    self.body.append(res)


def depart_tabbedstuff_html(self, node):
    divid = node["divid"]
    res = ""
    # close the tab plugin div and init the Bootstrap tabs
    res += END

    res = res % {"divid": divid}

    self.body.append(res)


class TabDirective(RunestoneDirective):
    """
    .. tab:: identifier
       :active: Optional flag that specifies this tab to be opened when page is loaded (default is first tab)--overridden by :inactive: flag on tabbedStuff

       Content
       ...


    config values (conf.py):

    - tabbed_div_class - custom CSS class of the component's outermost div
    """

    required_arguments = 1  # the name of the tab
    optional_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = RunestoneDirective.option_spec.copy()
    option_spec.update({"active": directives.flag})

    node_class = TabNode

    def run(self):
        """
        process the tab directive and generate html for output.
        :param self:
        :return:
        .. tab:: identifier
        :active: Optional flag that specifies this tab to be opened when page is loaded (default is first tab)--overridden by :inactive: flag on tabbedStuff

        Content
        ...
        """
        # Raise an error if the directive does not have contents.
        self.assert_has_content()

        # Create the node, to be populated by "nested_parse".
        self.options["tabname"] = self.arguments[0]
        tab_node = TabNode()
        tab_node["tabnode_options"] = self.options
        tab_node["tabname"] = self.options["tabname"]

        # Parse the child nodes (content of the tab)
        self.state.nested_parse(self.content, self.content_offset, tab_node)
        return [tab_node]


class TabbedStuffDirective(RunestoneDirective):
    """
    .. tabbed:: identifier
       :inactive: Optional flag that calls for no tabs to be open on page load

       Content (put tabs here)
       ...



    config values (conf.py):

    - tabbed_div_class - custom CSS class of the component's outermost div
    """

    required_arguments = 1  # the div to put the tabbed exhibit in
    optional_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = RunestoneDirective.option_spec.copy()
    option_spec.update({"inactive": directives.flag})

    def run(self):
        """
        process the tabbedStuff directive and generate html for output.
        :param self:
        :return:
        .. tabbed:: identifier
        :inactive: Optional flag that calls for no tabs to be open on page load

        Content (put tabs here)
        ...
        """

        # Raise an error if the directive does not have contents.
        self.assert_has_content()

        self.options["divid"] = self.arguments[0]
        self.validate_divid(self.options["divid"])

        env = self.state.document.settings.env
        self.options["divclass"] = env.config.tabbed_div_class

        # Create the node, to be populated by "nested_parse".
        tabbedstuff_node = TabbedStuffNode()
        tabbedstuff_node["tabbed_stuff_options"] = self.options
        tabbedstuff_node["divid"] = self.options["divid"]

        (
            tabbedstuff_node["source"],
            tabbedstuff_node["line"],
        ) = self.state_machine.get_source_and_line(self.lineno)

        # Parse the directive contents (should be 1 or more tab directives)
        self.state.nested_parse(self.content, self.content_offset, tabbedstuff_node)
        return [tabbedstuff_node]



# ***********************************************
# |docname| - Define the ``assignment`` directive
# ***********************************************
#
# Copyright (C) 2015  Paul Resnick
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import print_function

__author__ = "Paul Resnick"

from docutils import nodes
from docutils.parsers.rst import directives
from runestone.server.componentdb import (
    addAssignmentToDB,
    addAssignmentQuestionToDB,
    getCourseID,
    getQuestionID,
    get_HTML_from_DB,
)
from runestone.common.runestonedirective import (
    RunestoneDirective,
    RunestoneNode,
)
from datetime import datetime


def setup(app):
    app.add_directive("assignment", Assignment)
    app.add_node(AssignmentNode, html=(visit_a_html, depart_a_html))

    app.connect("doctree-resolved", process_nodes)
    app.connect("env-purge-doc", purge)


class AssignmentNode(nodes.General, nodes.Element, RunestoneNode):
    def __init__(self, content, **kwargs):
        """

        Arguments:
        - `self`:
        - `content`:
        """
        super(AssignmentNode, self).__init__(name=content["name"], **kwargs)
        self.a_components = content


def visit_a_html(self, node):
    pass


def depart_a_html(self, node):
    question_ids = node["a_components"]["question_ids"]
    basecourse = node["a_components"]["basecourse"]
    for q_id in question_ids:
        src = get_HTML_from_DB(q_id, basecourse)
        if src:
            self.body.append(src)
        else:
            self.body.append("<p>Missing HTML source for {}</p>".format(q_id))
            print(
                "No HTML source saved for {}; can't include that kind of question until code writing to HTML is implemented for that directive".format(
                    q_id
                )
            )


def process_nodes(app, env, docname):
    pass


def purge(app, env, docname):
    pass


class Assignment(RunestoneDirective):
    """
    .. assignment:
        :name: Problem Set 1
        :assignment_type: formative
        :questions: (divid_1 50), (divid_2 100), ...
        :deadline: 23-09-2016 15:30
        :points: integer
    """

    required_arguments = 0  # not a sphinx_id for these; just a name parameter
    optional_arguments = 0
    has_content = False
    option_spec = RunestoneDirective.option_spec.copy()
    option_spec.update(
        {
            "name": directives.unchanged,
            "assignment_type": directives.unchanged,
            "questions": directives.unchanged,
            "deadline": directives.unchanged,
            "points": directives.positive_int,
            "threshold": directives.positive_int,
            "autograde": directives.unchanged,
            "generate_html": directives.flag,
        }
    )

    def run(self):
        """
        .. assignment:
            :name: Problem Set 1
            :assignment_type: formative
            :questions: (divid_1 50), (divid_2 100), ...
            :deadline: 23-09-2016 15:30
            :points: integer
            :threshold: integer
            :autograde: visited
            :generate_html:
        """

        course_name = self.state.document.settings.env.config.html_context["course_id"]
        self.options["course_name"] = course_name

        ## No longer doing DB operations
        # course_id = getCourseID(course_name)
        # basecourse_name = self.state.document.settings.env.config.html_context.get('basecourse', "unknown")
        # self.options['basecourse'] = self.state.document.settings.env.config.html_context.get('basecourse', "unknown")

        name = self.options.get("name")  # required; error if missing

        unparsed = self.options.get("questions", None)
        question_names = []
        if unparsed:
            q_strings = unparsed.split(",")
            for q in q_strings:
                (question_name, points) = q.strip().split()
                question_names.append(question_name)
                # first get the question_id associated with question_name
                # question_id = getQuestionID(basecourse_name, question_name)
                # if question_id:
                #     addAssignmentQuestionToDB(question_id, assignment_id, points, autograde = autograde)
                # else:
                #     self.state.document.settings.env.warn(self.state.document.settings.env.docname, "Question {} is not in the database for basecourse {}".format(question_name, basecourse_name))
        else:
            self.state.document.settings.env.warn(
                self.state.document.settings.env.docname,
                "No questions for assignment {}".format(name),
            )
        self.options["question_ids"] = question_names

        if "generate_html" in self.options:
            assignment_node = AssignmentNode(self.options, rawsource=self.block_text)
            (
                assignment_node.source,
                assignment_node.line,
            ) = self.state_machine.get_source_and_line(self.lineno)
            return [assignment_node]
        else:
            return []



# .. Copyright (C) 2019 Bryan A. Jones.
#
# *************************************************************************
# |docname| - Definitions of shared data in the literate programming system
# *************************************************************************
# # This file contains a definition and supporting classes for a data structure used to export data from Sphinx to the Flask database for the books.
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8
# <http://www.python.org/dev/peps/pep-0008/#imports>`_.
#
# Standard library
# ----------------
import json
import os.path


# Given a file, return the inline comment token based on the file extension. Default to an empty string if we can't find the the provided language.
def commentForExt(file_name):
    return {
        ".c": "// ",
        ".css": "/* ",  # Will need some hand editing...
        ".h": "// ",
        ".ini": "; ",
        ".js": "// ",
        ".py": "# ",
        ".v": "// ",
        # Rust.
        ".rs": "// ",
        # PIC24 assembly
        ".s": "; ",
        # ARM assembly.
        ".S": "@ ",
    }.get(os.path.splitext(file_name)[1], "")


# Relative to the Sphinx output directory, the path to student source (which has all feedback and code solutions removed).
STUDENT_SOURCE_PATH = "../student_source"

# Relative to the Sphinx output directory, the path to the build system's output (which provides will be used to build and link student source).
BUILD_SYSTEM_PATH = "../build_system"

# Define a file in which Sphinx config values will be stored in the root directory of a Sphinx project.
SPHINX_CONFIG_NAME = "sphinx_settings.json"


def read_sphinx_config(
    # The directory in which the ``SPHINX_CONFIG_NAME`` file is located.
    dir_=".",
):

    try:
        with open(os.path.join(dir_, SPHINX_CONFIG_NAME), encoding="utf-8") as f:
            return json.loads(f.read())
    except IOError:
        return None


# Add the appropriate line comment delimiters to the ``CODE_HERE_STRING``.
def code_here_comment(file_name):
    return (
        _add_line_comment_delimiter(CODE_HERE_STRING, file_name)
        # Don't add a comment character for the trailing newline.
        + "\n"
    )


# A string which replaces code solutions in student source code.
CODE_HERE_STRING = """``*******************************************``
Add your code after this line.

Add your code before this line.
``*******************************************``"""


# Given a string and the file name it will be added to, add the line comment delimiter followed by a space to each line of the string.
def _add_line_comment_delimiter(
    # The string to which line comment delimiters should be added.
    str_,
    # The file name into which this string will be inserted.
    file_name,
):

    delim = commentForExt(file_name)
    # Add a comment for to every newline.
    str_ = str_.replace("\n", "\n" + delim)
    # Begin the string with a comment as well.
    return delim + str_



# extract text from html format
def extractText(blob):
    # checking for inner <code> tag
    if "<code" in blob:
        return extractTextHelper("", blob)

    return blob


def extractTextHelper(title, blob):
    if "<code" not in blob:
        title += blob
        return title

    idx0 = blob.find("<code")

    firstSub = blob[0:idx0]
    title += firstSub

    idx1 = blob.find('"pre">')
    idx2 = blob.find("</span>")
    secondSub = blob[idx1 + 6 : idx2]
    title += secondSub
    title += " "

    idx3 = blob.find("</code>")

    # checking for inner <span> tag
    spareBlob = blob[idx2 + 7 : idx3]

    if "<span" in spareBlob:
        title = spareBlobHelper(title, spareBlob)

    thirdSub = blob[idx3 + 7 :]

    return extractTextHelper(title, thirdSub)


def spareBlobHelper(title, spareBlob):
    if "<span" not in spareBlob:
        return title

    spareIdx0 = spareBlob.find('"pre">')
    spareIdx1 = spareBlob.find("</span>")

    spareSub = spareBlob[spareIdx0 + 6 : spareIdx1]

    title += spareSub
    title += " "

    spareBlob = spareBlob[spareIdx1 + 7 :]

    return spareBlobHelper(title, spareBlob)


def extractTextII(blob):
    # checking for <strong> tag
    if "</span>" in blob:
        blob = blob.replace('<span class="section-number">',"")
        blob = blob.replace("</span>","")
        return blob
    if "<strong>" in blob:
        title = ""
        idx0 = blob.find("<strong>")

        firstSub = blob[0:idx0]
        title += firstSub

        idx1 = blob.find("</strong")
        secondSub = blob[idx0 + 8 : idx1]
        title += secondSub
        title += " "

        thirdSub = blob[idx1 + 10 :]
        title += thirdSub

        return title

    return blob


def setup(app):
    # adding custom filter functions to jinja2's global FILTERS dictionary
    import jinja2

    jinja2.filters.FILTERS["extractText"] = extractText
    jinja2.filters.FILTERS["extractTextII"] = extractTextII



# ************************************************************
# |docname| -- Define validation for endpoint query parameters
# ************************************************************
# This file contains the models we use for post requests and for type checking throughout the application.
# These object models should be used wherever possible to ensure consistency

# Imports
# =======
# These are listed in the order prescribed by `PEP 8`_.
#
# Standard library
# ----------------
from datetime import datetime
from dateutil.parser import isoparse
from typing import Container, Optional, Type, Dict, Tuple, Any, Union, List
# Third-party imports
# -------------------
from pydantic import (
    field_validator,
    StringConstraints,
    ConfigDict,
    BaseModel,
    BaseConfig,
    create_model,
    Field,
    Json,
)
from humps import camelize  # type: ignore
from typing_extensions import Annotated

# Local application imports
# -------------------------
# None.
from rsptx.response_helpers.core import canonical_utcnow

# Schema generation
# =================
# Change the ``BaseModel.from_orm`` method to return ``None`` if the input was ``None``, instead of a class full of variables set to ``None``.
class BaseModelNone(BaseModel):
    @classmethod
    def from_orm(cls, obj):
        return None if obj is None else super().from_orm(obj)

    model_config = ConfigDict(from_attributes=True)


# This creates then returns a Pydantic schema from a SQLAlchemy Table or ORM class.
#
# This is copied from https://github.com/tiangolo/pydantic-sqlalchemy/blob/master/pydantic_sqlalchemy/main.py then lightly modified.
def sqlalchemy_to_pydantic(
    # The SQLAlchemy model -- either a Table object or a class derived from a declarative base.
    db_model: Type,
    *,
    # An optional Pydantic `model config <https://pydantic-docs.helpmanual.io/usage/model_config/>`_ class to embed in the resulting schema.
    config: Optional[Type[BaseConfig]] = None,
    # The base class from which the Pydantic model will inherit.
    base: Type[BaseModel] = BaseModelNone,
    # SQLAlchemy fields to exclude from the resulting schema, provided as a sequence of field names. Ignore the id field by default.
    exclude: Container[str] = tuple(),
) -> Type[BaseModel]:
    # If provided an ORM model, get the underlying Table object.
    db_model = getattr(db_model, "__table__", db_model)

    fields: Dict[str, Union[Tuple[Type, Any], Type[BaseConfig]]] = {}
    for column in db_model.columns:
        # Determine the name of this column.
        name = column.key
        if name in exclude:
            continue

        # Determine the Python type of the column.
        python_type = column.type.python_type
        if python_type == str and hasattr(column.type, "length"):
            python_type = Annotated[
                str, StringConstraints(max_length=column.type.length)
            ]

        # Determine if the column can be null, meaning it's optional from a Pydantic perspective. Make the id column optional, since it won't be present when inserting values to the database.
        if column.nullable or name == "id":
            python_type = Optional[python_type]

        # Determine the default value for the column. Allow the id column to be null.
        default = column.default
        if callable(default):
            default = column.default()
        if column.default is None and not column.nullable and name != "id":
            default = ...

        # Build the schema based on this info.
        fields[name] = (python_type, default)

    # See `create_model <https://pydantic-docs.helpmanual.io/usage/models/#dynamic-model-creation>`_.
    if config:
        fields["__config__"] = config
    pydantic_model = create_model(str(db_model.name), __base__=base, **fields)  # type: ignore
    return pydantic_model


# Schemas
# =======
class LogItemIncoming(BaseModelNone):
    """
    This class defines the schema for what we can expect to get from a logging event.
    Because we are using pydantic type verification happens automatically, if we want
    to add additional constraints we can do so.
    """

    event: str
    act: str
    div_id: str
    course_name: str
    sid: Optional[str] = None
    answer: Optional[str] = None
    correct: Optional[Union[bool, int]] = None
    percent: Optional[float] = None
    clientLoginStatus: Optional[bool] = None
    timezoneoffset: Optional[float] = None
    timestamp: Optional[datetime] = None
    chapter: Optional[str] = None
    subchapter: Optional[str] = None
    assignment_id: Optional[int] = None
    # used by parsons
    source: Optional[str] = None
    # used by dnd
    min_height: Optional[int] = None
    # used by unittest
    passed: Optional[int] = None
    failed: Optional[int] = None
    # used by timed exam
    incorrect: Optional[int] = None
    skipped: Optional[int] = None
    time_taken: Optional[int] = None


class AssessmentRequest(BaseModelNone):
    course: str
    div_id: str
    event: str
    sid: Optional[str] = None
    # See `Field with dynamic default value <https://pydantic-docs.helpmanual.io/usage/models/#required-optional-fields>`_.
    deadline: datetime = Field(default_factory=canonical_utcnow)
    # TODO[pydantic]: The following keys were removed: `json_encoders`.
    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-config for more information.
    model_config = ConfigDict(
        json_encoders={
            datetime: lambda v: v.isoformat(),
        }
    )

    @field_validator("deadline", mode="before")
    @classmethod
    def time_validate(cls, v):
        if v:
            return isoparse(v)
        return isoparse(datetime.isoformat(canonical_utcnow()))


class TimezoneRequest(BaseModelNone):
    timezoneoffset: float


class LogRunIncoming(BaseModelNone):
    div_id: str
    code: str
    errinfo: str
    to_save: bool
    course: str
    clientLoginStatus: bool
    timezoneoffset: float
    language: str
    prefix: Optional[str] = None
    suffix: Optional[str] = None
    partner: Optional[str] = None
    sid: Optional[str] = None
    editDist: Optional[int] = None
    changesPerSecond: Optional[float] = None


# Schemas for Completion Data
# ---------------------------
class LastPageDataIncoming(BaseModelNone):
    last_page_url: str  # = Field(None, alias="lastPageUrl") is the manual way
    course_id: str = Field(alias="course")
    completion_flag: int
    pageLoad: bool
    markingComplete: bool
    markingIncomplete: bool
    last_page_scroll_location: int
    is_ptx_book: bool
    model_config = ConfigDict(alias_generator=camelize)


class LastPageData(BaseModelNone):
    last_page_url: str
    course_name: str = Field(alias="course_id")
    completion_flag: int
    last_page_scroll_location: int
    last_page_chapter: str
    last_page_subchapter: str
    last_page_accessed_on: datetime
    user_id: int


class SelectQRequest(BaseModel):
    selector_id: str
    questions: Optional[str] = None
    proficiency: Optional[str] = None
    points: Optional[int] = None
    min_difficulty: Optional[float] = None
    max_difficulty: Optional[float] = None
    not_seen_ever: Optional[bool] = None
    autogradable: Optional[bool] = None
    primary: Optional[bool] = None
    AB: Optional[str] = None
    toggleOptions: Optional[str] = None
    timedWrapper: Optional[str] = None
    limitBaseCourse: Optional[str] = None


class PeerMessage(BaseModel):
    type: str
    sender: str
    message: str
    broadcast: bool


class AssignmentIncoming(BaseModel):
    name: str
    description: str
    points: int
    duedate: datetime
    kind: str


class QuestionIncoming(BaseModel):
    id: Optional[int] = None
    name: str
    source: str
    question_type: str
    htmlsrc: str
    autograde: Optional[str] = None
    question_json: Json
    chapter: Optional[str] = None
    author: Optional[str] = None
    tags: Optional[str] = None
    description: Optional[str] = None
    difficulty: Optional[float] = None
    topic: Optional[str] = None


class AssignmentQuestionIncoming(BaseModel):
    assignment_id: int
    question_id: int
    points: int
    order: Optional[int] = None


class SearchSpecification(BaseModel):
    source_regex: Optional[str] = None
    question_type: Optional[str] = None
    author: Optional[str] = None
    tag_list: Optional[str] = None
    base_course: Optional[str] = None


class ScoringSpecification(BaseModel):
    assigned: bool = False
    score: Optional[Union[int, float]] = None
    max_score: Optional[int] = None
    question_id: Optional[int] = None
    assignment_id: Optional[int] = None
    which_to_grade: Optional[str] = None
    how_to_score: Optional[str] = None
    username: Optional[str] = None
    comment: Optional[str] = None


class ReadingAssignmentSpec(BaseModel):
    activities_required: int
    question_id: int
    assignment_id: int
    points: int
    name: str

class UpdateAssignmentExercisesPayload(BaseModel):
    isReading: bool
    assignmentId: int
    idsToAdd: Optional[List[int]] = None
    idsToRemove: Optional[List[int]] = None



from enum import Enum
from rsptx.data_types.question_type import QuestionType

class WhichToGradeOptions(Enum):
    FIRST_ANSWER = ("first_answer", "First answer", QuestionType.all_types())
    LAST_ANSWER = ("last_answer", "Last answer", QuestionType.all_types())
    BEST_ANSWER = ("best_answer", "Best answer", QuestionType.all_types())
    ALL_ANSWER = ("all_answer", "All peer votes", [QuestionType.MCHOICE])

    def to_dict(self):
        return {
            "value": self.value[0],
            "label": self.value[1],
            "supported_question_types": [qt.value_only() for qt in self.value[2]],
        }
